import os
import sys


def read_msr(index):
    """
    Read a 'index' register from MSR.
    Return tuple of two integer numbers: first represents 31:0 bits,
    second represents 63:32 bits.
    """

    import struct

    try:
        with open("/dev/cpu/0/msr", 'r', 0) as src:
            src.seek(index)

            val = struct.unpack('Q', src.read(8))[0]
            return (val & 0xffffffff, val >> 32)

    except Exception, err:
        print >> sys.stderr, "Failed to read MSR: ", repr(err)
        return 0, 0


class CPU(object):
    """
    Represents generic CPU, provide useful contants and access to MSR.
    """

    def __init__(self):
        self.MSR_IA32_VMX_PROCBASED_CTLS2 = 0x48b

        self.features = []

        with open("/proc/cpuinfo") as src:
            for line in src.readlines():
                if not line.startswith("flags"):
                    continue

                # Skip 'flags :'
                self.features = line.split()[2:]
                break

    def has_feature(self, name):
        """
        Check whether a 'name' feature is supported by the CPU.
        (i.e. exists in /proc/cpuinfo flags).
        """

        return name in self.features

    @staticmethod
    def get():
        """
        Return object representing current CPU. The object's type is
        determined by vendor information extracted from /proc/cpuinfo.
        """

        with open("/proc/cpuinfo") as src:
            for line in src.readlines():
                if not line.startswith('vendor_id'):
                    continue

                if "GenuineIntel" in line:
                    return Intel()

                if "AuthenticAMD" in line:
                    return AMD()

                # Unsupported vendor
                return None

        return None

    def kvm_loaded(self):
        return os.system("/usr/sbin/modprobe kvm-%s" % str.lower(self.__class__.__name__)) == 0

class Intel(CPU):
    """
    Represents Intel processors. Allow to check support of hardware
    virtualization (VT-x) and required CPU features using /proc/cpuinfo and
    Model-Specific Registers (MSR) as a data source.
    For additional details please refer to 'Intel Software Developer's Manual
    Volume 3B: System Programming Guide, Part 2'.
    """

    def vt_enabled(self):
        """
        Check support of basic virtualization technology
        (VMX flag in /proc/cpuinfo).
        """

        return self.has_feature("vmx")

    def full_featured(self):
        """
        Check support of 'Extended Page Tables' (EPT flag in /proc/cpuinfo)
        and 'Unrestricted guest' (32+7 bit in MSR_IA32_VMX_PROCBASED_CTLS2
        control) features support.
        """

        val = read_msr(self.MSR_IA32_VMX_PROCBASED_CTLS2)[1]
        return self.has_feature("ept") and bool(val & (1 << 7))

    def __str__(self):
        return os.linesep.join([
            "Intel processor:",
            "\tVMX: %s" % self.vt_enabled(),
            "\tEPT and Unrestricted guest: %s" % self.full_featured()
        ])


class AMD(CPU):
    """
    Represents AMD processors. Allow to check support of hardware
    virtualization (SVM) and required CPU features using using /proc/cpuinfo
    and Model-Specific Registers (MSR) as a data source.
    For additional details please refer to 'AMD64 Architecture
    Programmer's Manual Volume 2: System Programming'.
    """

    def vt_enabled(self):
        """
        Check support of basic virtualization technology
        (SVM flag in /proc/cpuinfo).
        """

        return self.has_feature("svm")

    def full_featured(self):
        """
        Check support of 'Rapid Virtualization Indexing'
        (RVI) aka 'Nested Page Tables' (NPT flag in /proc/cpuinfo).
        """

        return self.has_feature("npt")

    def __str__(self):
        return os.linesep.join([
            "AMD processor:",
            "\tSVM: %s" % self.vt_enabled(),
            "\tRVI: %s" % self.full_featured()
        ])


def main():
    """
    Show general information about virtualization support
    of the CPU in the system.
    """

    print >> sys.stdout, CPU.get()
    return 0

if __name__ == "__main__":
    sys.exit(main())
