#
# virtuozzo.py
#
# Copyright (c) 2015-2020 Virtuozzo International GmbH. All rights reserved.
#
# This file is part of OpenVZ. OpenVZ is free software;
# you can redistribute it and/or modify it under the terms of the GNU
# Lesser General Public License as published by the Free Software Foundation;
# either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
# Schaffhausen, Switzerland.

from pyanaconda.installclass import BaseInstallClass
from pyanaconda.constants import *
from pyanaconda.product import productName
from pyanaconda.flags import flags
from pyanaconda import network
from pyanaconda import nm
from pyanaconda.i18n import _
from pyanaconda.iutil import getSysroot
from pyanaconda.virtuozzo_c import in_cmdline
from pyanaconda.packaging.yumpayload import YumPayload
from pykickstart.constants import CLEARPART_TYPE_ALL
from blivet.partspec import PartSpec
from blivet.platform import platform
from blivet.devicelibs import swap
from blivet.size import Size

import os, types, time, re

class VirtuozzoBaseInstallClass(BaseInstallClass):
    # name has underscore used for mnemonics, strip if you dont need it
    vendor = "Virtuozzo"
    id = "virtuozzo"
    name = N_("%s" % vendor)
    sortPriority = 20000
    if productName != vendor and productName != vendor + " Hybrid Server" and productName != "OpenVZ":
        hidden = True

    def __init__(self):
        BaseInstallClass.__init__(self)

        self.is_vz = True
        self.agent_release = "va-agent-release"
        self.create_bridges = True
        self.fix_slave_cfgs = True
        self.fix_ifcfgs = False
        self.defaultFS = "ext4"
        self.bootloaderTimeoutDefault = 5
        self.installUpdates = True
        self._l10n_domain = "comps"
        self.efi_dir = "virtuozzo"
        self.defostemplate = "vzlinux-7-x86_64"
        self.ostemplate_cache_suffix = "plain.ploopv2.tar.lz4"
        self.dcachepath = "/vz/template/cache"
        self.isocachedir = "vztcache"

        # ISO repodir
        self.iso_repo = MOUNT_DIR + "/repo"
        self.va_agent_repo = "va-agent"
        self.va_mn_repo = "va-mn"
        self.net_va_mn_mirrors = ["http://repo.virtuozzo.com/va-mn/mirrorlists/7.0/releases-os.mirrorlist",
            "http://repo.virtuozzo.com/va-mn/mirrorlists/7.0/updates-os.mirrorlist"]
        self.net_va_agent_mirrors = ["http://repo.virtuozzo.com/va-agent/mirrorlists/7.0/releases-os.mirrorlist",
            "http://repo.virtuozzo.com/va-agent/mirrorlists/7.0/updates-os.mirrorlist"]
        self.net_base_mirrors = ["http://repo.virtuozzo.com/vzlinux/mirrorlist/mirrors-7-os",
            "http://repo.virtuozzo.com/vzlinux/mirrorlist/mirrors-7-updates",
            "http://repo.virtuozzo.com/vz/mirrorlists/7.0/releases-os.mirrorlist",
            "http://repo.virtuozzo.com/vz/mirrorlists/7.0/updates-os.mirrorlist"]

        self.base_repos = []
        self.base_mlists = []
        self.base_va_agent_repos = []
        self.base_va_agent_mlists = []
        self.base_va_mn_repos = []
        self.base_va_mn_mlists = []

        # up2date repos
        self.up2date_repos = []

        # Dispatcher first start scripts
        self.prl_disp_first_start_script_non_chroot = "/usr/libexec/prl_disp_first_start"
        self.prl_disp_first_start_script = getSysroot() + self.prl_disp_first_start_script_non_chroot
        self.prl_disp_first_start_header = "#!/bin/bash\n\n" \
            "mkdir -p /etc/vz\n" \
            "touch /etc/vz/prl_disp_disable_network_sync\n\n"
        self.prl_disp_first_start_footer = "# Cleanup\n" \
            "rm -f %s\n" \
            "exit 0\n" % self.prl_disp_first_start_script_non_chroot
        self.prl_disp_first_start_lines = []

        self.select_minsize_slash = True
        self.efiboot_size = Size("200MiB")
        self.biosboot_size = Size("1MiB")
        self.boot_size = Size("1GiB")
        self.slash_size = Size("12GiB")
        self.max_slash_size = Size("64GiB")
        self.min_ssd_size = Size("30GiB") # With /boot we require >= 32G SSDs
        self.min_vz_size = Size("30GiB")
        self.min_cs_size = Size("100GB")
        self.swap_size = swap.swapSuggestion(smallswap=("smallswap" in flags.cmdline))
        self.anaconda = None
        self.mask_nm = False
        self.docker_img_ver = ""
        self.reboot_prohibited = "/tmp/reboot_prohibited"
        self.css = ""
        self.simplify_network = False
        self.prohibited_net_devs = ()
        self.ntp_always_on = False
        self.step_by_step_spokes = []
        self.default_font = "OpenSans"
        self.show_rnotes = True
        self.updates_repodir = "/tmp/yum.update.repos"

    @property
    def rnt_rescue_image(self):
        return False

    @property
    def not_ask_reboot(self):
        return False

    @property
    def bootreq_size(self):
        return self.efiboot_size + self.biosboot_size + self.boot_size

    @property
    def min_hdd_size(self):
        return self.swap_size + self.min_vz_size + self.slash_size

    def prl_disp_first_start(self, lines):
        self.prl_disp_first_start_lines += lines
        with open(self.prl_disp_first_start_script, "w") as f:
            f.write(self.prl_disp_first_start_header)
            for line in self.prl_disp_first_start_lines:
                f.write(line + "\n")
            f.write(self.prl_disp_first_start_footer)
            f.close()
            os.chmod(self.prl_disp_first_start_script, 0755)

    def setDefaultPartitioning(self, storage):
        # Fill here only basic partitions for compatibility
        autorequests = [PartSpec(mountpoint="/", fstype="ext4",
                                 size=Size("20GiB"), maxSize=Size("50GiB"), grow=True,
                                 requiredSpace=Size("20GiB"), btr=False, lv=True, encrypted=False)]

        bootreqs = platform.setDefaultPartitioning()
        if bootreqs:
            autorequests.extend(bootreqs)

        swp = swap.swapSuggestion(smallswap=("smallswap" in flags.cmdline))
        autorequests.append(PartSpec(fstype="swap", size=swp, grow=False,
                                     lv=True, encrypted=True))

        for autoreq in autorequests:
            if autoreq.fstype is None:
                if autoreq.mountpoint == "/boot":
                    autoreq.fstype = storage.defaultBootFSType
                else:
                    autoreq.fstype = storage.defaultFSType

        storage.autoPartitionRequests = autorequests

    def configure(self, anaconda):
        BaseInstallClass.configure(self, anaconda)
        # Check for kickstart autopart - prohibit without clearpart --all
        if flags.automatedInstall and anaconda.ksdata.autopart.autopart and \
                anaconda.ksdata.clearpart.type != CLEARPART_TYPE_ALL or \
                len(anaconda.ksdata.clearpart.devices) > 0:
            raise RuntimeError("The autopart kickstart key is only used with " \
                "clearpart --all and without --list, --linux, or --none")
        anaconda.storage.vstorage_cs = anaconda.ksdata.vstorage.cs_join
        self.setDefaultPartitioning(anaconda.storage)
        # Enabled update repos for yum
        if anaconda.ksdata.up2date.enabled and isinstance(anaconda._payload, YumPayload):
            for repo in anaconda._payload._yum.repos.repos.values():
                if not repo.id.endswith("-updates"):
                    continue
                self.up2date_repos.append(repo.id)
                # Also, enable main repos
                mrepo = re.sub("-updates$", "", repo.id)
                for mainrepo in (mrepo, mrepo + "-base", mrepo + "-os"):
                    if mainrepo not in (r.id for r in anaconda._payload._yum.repos.repos.values()):
                        continue
                    self.up2date_repos.append(mainrepo)
                    break
        self.anaconda = anaconda

    def setNetworkOnbootDefault(self, ksdata):
        if flags.automatedInstall:
            return
        for dev in nm.nm_devices():
            try:
                if nm.nm_device_type_is_wifi(dev):
                    continue
            except:
                continue
            try:
                network.update_onboot_value(dev, dev in nm.nm_activated_devices(), ksdata=ksdata)
            except:
                pass

    @property
    def is_free(self):
        return productName == "OpenVZ"

    @property
    def get_release_f(self):
        return "/etc/system-release"

    @property
    def get_eula(self):
        if self.is_free:
            return "/usr/share/licenses/openvz-release/NOTICE"
        return "/usr/share/licenses/virtuozzo-release/EULA"

    @property
    def is_squashroot(self):
        if not os.path.exists("/squashroot") or not self.anaconda:
            return False
        if in_cmdline("nosquashroot"):
            return False
        return self.anaconda.ksdata.squashinstall.enabled

    def reboot(self):
        while os.path.exists(self.reboot_prohibited):
            time.sleep(5)

    def install_updates(self):
        return
