from pyanaconda.flags import flags
from pyanaconda.iutil import execInSysroot, execWithRedirect, getSysroot, execWithCapture, ipmi_report, stop_service, expert_mode
from pyanaconda.product import productVersion
from pyanaconda.network import process_opensm
from pyanaconda import hvt_checker
from pyanaconda.ui.lib.disks import getDisks
from pyanaconda.threads import threadMgr, AnacondaThread
from pyanaconda.packaging.yumpayload import YumPayload
from pyanaconda.packaging import PayloadInstallError
from pyanaconda.errors import errorHandler, ERROR_RAISE
from pyanaconda.constants import BASE_REPO_NAME, THREAD_WAIT_FOR_CONNECTING_NM, THREAD_INSTALL
from pyanaconda.i18n import _
from pykickstart.constants import CLEARPART_TYPE_ALL
from pyanaconda.progress import progressQ, progress_report
from pyanaconda.ui.communication import hubQ
from pyanaconda.nm import nm_devices, nm_device_ip_config, nm_activated_devices
from pyanaconda.threads import threadMgr, AnacondaThread
from blivet.devicelibs.mpath import is_multipath_member
from blivet import udev
from netaddr import IPNetwork
import storage_utils
import logging
import urlgrabber
import shutil
import glob
import dbus
log = logging.getLogger("anaconda")
from blivet.errors import PartitioningError, NotEnoughFreeSpaceError
from blivet.size import Size
from blivet.devicelibs import swap
from blivet.platform import platform
from blivet.partitioning import growLVM, doPartitioning
from blivet.formats import getFormat

vstorage_bin = "/usr/bin/vstorage"
THREAD_VSTORAGE_WAIT_STORAGE_INIT = "VstorageWaitStorageInitThread"

import re
import os
import time
import sys

class DisksUseCaps(object):
    def __init__(self, disks):
        self.disks = disks
        self.disks_vz = {}
        self.fill()

    def fill(self):
        for d in self.disks:
            self.disks_vz[d.name] = {}
            for t in ("boot", "vzdata", "slash", "csdata"):
                for s in ("cap", "use"):
                    exec("self.disks_vz[d.name][\"%s_%s\"] = d.%s_%s" % (s, t, s, t))

    def restore(self, disks):
        for d in disks:
            for v in self.disks_vz:
                if d.name != v:
                    continue
                for s in self.disks_vz[v]:
                    exec("d.%s = self.disks_vz[v][s]" % s)

    def __eq__(self, y):
        if not y:
            log.debug("DisksUseCaps: null")
            return False

        for v in self.disks_vz:
            catched = False
            for w in y.disks_vz:
                if w != v:
                    continue
                catched = True
                for s in self.disks_vz[v]:
                    for c in y.disks_vz[w]:
                        if c != s:
                            continue
                        if self.disks_vz[v][s] != y.disks_vz[w][c]:
                            log.debug("DisksUseCaps: %s.%s != %s.%s" % (v, s ,w ,c))
                            return False
            if not catched:
                log.debug("DisksUseCaps: not cached")
                return False

        return True

    def __ne__(self, y):
        return not self == y

class MDdev(object):
    def __init__(self, storage, disks, fmt_type, instclass, size = None, mountpoint = None,
            weight = 0, level = 1, grow = False, maxsize = None):
        self.level = level
        self.grow = grow
        self.maxsize = maxsize
        self.storage = storage
        self.disks = disks
        self.instclass = instclass
        self.size = size
        self.fmt_type = fmt_type
        self.mountpoint = mountpoint
        self.weight = weight
        self.parts = []
        self.device = None
        self.partition_table_size = Size("2M") # 1M at the start + 1M at the end of HDD
        if not disks:
            raise PartitioningError(_("No disks available"))

    def disk_size_wo_mandatory_parts(self, size):
        return size - (self.instclass.bootreq_size +
            self.instclass.swap_size + self.partition_table_size)

    def align_maxsize(self):
        if not self.grow:
            return
        # 0 means maximum available, so drop it to real maximum
        if self.maxsize == 0:
            for disk in self.disks:
                self.maxsize = max(self.maxsize,
                        self.disk_size_wo_mandatory_parts(disk.size))
        for disk in self.disks:
            self.maxsize = min(self.maxsize,
                    self.disk_size_wo_mandatory_parts(disk.size))

    def create_members(self):
        # Fallback to basic partition when one drive available
        if len(self.disks) == 1:
            self.device = self.storage.newPartition(fmt_type=self.fmt_type,
                                            size=self.size,
                                            mountpoint=self.mountpoint,
                                            parents=self.disks,
                                            weight=self.weight,
                                            grow=self.grow,
                                            maxsize=self.maxsize)
            self.storage.createDevice(self.device)
            return
        # Recalculate maxsize as maximum HDD available size
        self.align_maxsize()
        for disk in self.disks:
            part = self.storage.newPartition(fmt_type="mdmember",
                                            size=self.size,
                                            parents=[disk],
                                            weight=self.weight,
                                            grow=self.grow,
                                            maxsize=self.maxsize)
            self.storage.createDevice(part)
            self.parts.append(part)

    def create_md(self):
        if len(self.disks) == 1:
            return
        self.device = self.storage.newMDArray(fmt_type=self.fmt_type,
                                            memberDevices=len(self.parts),
                                            totalDevices=len(self.parts),
                                            level=self.level,
                                            mountpoint=self.mountpoint,
                                            parents=self.parts)
        self.storage.createDevice(self.device)

class DoVZAutopartPreselect(object):
    def __init__(self, instClass, storage, ksdata):
        self.instClass = instClass
        self.storage = storage
        self.ksdata = ksdata
        self.reservedSize = Size("8MiB")
        self.slash_drive = None
        self.hdds = []
        self.disks = getDisks(storage.devicetree)

    def check_min_hdd_size(self, size):
        # Function that checks for disk with required minimum size available
        return len([ d for d in self.hdds if d.availsize >= size ]) != 0

    def reset_drives_selection(self):
        for disk in self.disks:
            disk.is_ssd = disk.use_ssd = False
            disk.cap_boot = disk.use_boot = False
            disk.cap_slash = disk.use_slash = False
            disk.cap_csdata = disk.use_csdata = False
            disk.cap_vzdata = disk.use_vzdata = False

        self.preselect_drives()

    def set_csdata_non_slash_drives(self):
        # Vstorage case - vzdata on slash, hdds mark as CS used
        for disk in [ d for d in self.hdds if d.cap_csdata and
                not d.use_slash ]:
            disk.use_csdata = True
            log.info("%s drive selected for CS" % disk.name)
            disk.use_boot = False
            log.info("%s drive deselected for boot" % disk.name)

    def preselect_vstorage(self):
        if not self.instClass.is_vz:
            return
        self.set_csdata_non_slash_drives()
        # It's safe because of we skip HDDs which not capable for / + /vz
        self.slash_drive.use_vzdata = True
        log.info("%s drive selected for VZ" % self.slash_drive.name)
        self.slash_drive.availsize -= self.instClass.min_vz_size
        # Try to place CS together with slash
        if not [ d for d in self.disks if d.use_csdata ] and \
            self.slash_drive.availsize > self.instClass.min_cs_size:
                self.slash_drive.use_csdata = True
                log.info("%s slash_drive selected for CS" % self.slash_drive.name)

    def set_vzdata_non_slash_drives(self):
        # Non-vstorage - mark vzdata capable drives
        for disk in [ d for d in self.hdds if d.cap_vzdata and
                not d.use_slash ]:
            disk.use_vzdata = True
            log.info("%s drive selected for VZ" % disk.name)


    def preselect_non_vstorage(self):
        if not self.instClass.is_vz:
            return
        # For kickstart in 'ui' mode preselect only slash drive for /vz
        if not flags.automatedInstall or not 'ui' in flags.cmdline:
            self.set_vzdata_non_slash_drives()
        if not [ d for d in self.disks if d.use_vzdata ]:
            self.slash_drive.use_vzdata = True
            log.info("%s slash_drive selected for VZ" % self.slash_drive.name)

    def preselect_drives(self):
        # First - select separatelly drives for SSD and boot
        for disk in self.disks:
            # Check for protected - and exclude such devices
            # (e.g. installation flash for now)
            # at early stage
            if disk.protected:
                log.info("Skipped protected device %s" % disk.name)
                continue

            # Also skip all other USB HDDs/flashes
            if not in_cmdline("allow_usb_hdd") and re.match(".*\/usb.*", disk.sysfsPath):
                log.info("Skipped USB device %s" % disk.name)
                disk.protected = True
                continue

            if not in_cmdline("allow_mpath") and (is_multipath_member(disk.path) or \
                    udev.device_is_dm_mpath(udev.get_device(disk.sysfsPath))):
                log.info("Skipped mpath device %s" % disk.name)
                disk.protected = True
                continue

            # Set availsize - will use later
            # Reserve size for LVM metadata
            disk.availsize = disk.size - self.reservedSize

            # Skip not selected drives - only kickstart case
            if (flags.automatedInstall and self.ksdata.ignoredisk.onlyuse and not \
                    disk.name in self.ksdata.ignoredisk.onlyuse) or \
                    disk.name in self.ksdata.ignoredisk.ignoredisk:
                log.info("Skipped drive %s during autopartitioning: drive not selected" % disk.name)
                continue

            # Skip too small drives for boot
            if disk.size < self.instClass.bootreq_size:
                log.info("Skipped drive %s during autopartitioning: too small even for " \
                    "bootreq - %s" % (disk.name, disk.size))
                continue

            # Select drives for boot
            disk.availsize -= self.instClass.bootreq_size
            disk.cap_boot = True
            log.info("%s drive selected for boot" % disk.name)

            # Select drives for SSD cache
            if is_ssd(disk):
                disk.is_ssd = disk.use_ssd = True
                if disk.availsize >= self.instClass.min_hdd_size:
                    disk.cap_slash = True
                    disk.cap_vzdata = True
                if disk.availsize >= self.instClass.min_hdd_size + self.instClass.min_cs_size:
                    disk.cap_csdata = True
                log.info("%s drive selected for SSD cache" % disk.name)
                continue

            self.hdds.append(disk)

        # Try to find non-SSD drive enough to minimal swap + / + vz
        if not self.check_min_hdd_size(self.instClass.min_hdd_size):
            for disk in self.disks:
                disk.use_ssd = False
                self.hdds.append(disk)
            if not self.check_min_hdd_size(self.instClass.min_hdd_size):
                # Here we fail with autopart: BIP-BIP-BIP happens, too low space
                # Let user partition this BIP-BIP-BIP manually
                log.error("Too small size for autopart, autopart disabled")
                return

        self.storage.can_autopart = True
        log.info("Autopartition supported")

        # Select drive for / and swap
        slash_capable_hdds = [ d for d in self.hdds if d.availsize >= self.instClass.min_hdd_size ]
        for disk in slash_capable_hdds:
            disk.cap_slash = True
            log.info("%s drive is slash capable" % disk.name)
        if self.instClass.select_minsize_slash:
            self.slash_drive = sorted(slash_capable_hdds, key = lambda k: k.size)[0]
        else:
            self.slash_drive = sorted(slash_capable_hdds, key = lambda k: k.name)[0]
        self.slash_drive.use_slash = self.slash_drive.use_boot = True
        log.info("%s drive selected for slash" % self.slash_drive.name)
        self.slash_drive.availsize -= (self.instClass.swap_size + self.instClass.slash_size)

        # Next, check for drives, which can be used for VZ
        for disk in [ d for d in self.hdds if d.availsize >= self.instClass.min_vz_size ]:
            disk.cap_vzdata = True
            log.info("%s drive is vzdata capable" % disk.name)

        # Next, check for drives, which can be used for CSes. Do not calculate reserved and bootreq
        # them are too small relative to min_cs_size
        for disk in [ d for d in self.hdds if d.availsize >= (self.instClass.min_cs_size - \
            self.reservedSize - self.instClass.bootreq_size) ]:
            disk.cap_csdata = True
            log.info("%s drive is csdata capable" % disk.name)

        # Next mark drives for data
        if self.storage.vstorage_cs:
            self.preselect_vstorage()
            return

        self.preselect_non_vstorage()


class DoVZAutopart(object):
    def __init__(self, instClass, storage):
        self.instClass = instClass
        self.storage = storage
        self.mddevs = []
        self.slash_mddevs = []
        self.vg = []
        self.csnum = 0
        self.ssdnum = 0

    @property
    def get_boot_disks(self):
        return [ d for d in self.storage.disks if d.use_boot ]

    def prepare_boot(self, boot_disks):
        from blivet.platform import EFI, MacEFI
        # Create RAID1 on every disk for boot
        # 1MiB biosboot reserve or EFI + /boot
        if isinstance(platform, EFI):
            fmt_type = "efi"
            if isinstance(platform, MacEFI):
                fmt_type = "macefi"
            self.mddevs.append(MDdev(self.storage, boot_disks, fmt_type, self.instClass,
                size=self.instClass.efiboot_size, mountpoint="/boot/efi", weight=2000))
            log.info("Created /boot/efi instance on %s" % ', '.join(d.name for d in boot_disks))
        else:
            # Create biosboot required by GPT on arch HDD
            for disk in [ d for d in boot_disks if getattr(d.format, "labelType", None) == "gpt" ]:
                self.storage.createDevice(self.storage.newPartition(fmt_type="biosboot",
                                                size=self.instClass.biosboot_size,
                                                parents=[disk],
                                                weight=5000))
                log.info("Created biosboot instance on %s" % disk.name)

        self.mddevs.append(MDdev(self.storage, boot_disks, "ext4", self.instClass,
            size=self.instClass.boot_size, mountpoint="/boot", weight=2000))
        log.info("Created /boot instance on %s" % ', '.join(d.name for d in boot_disks))

    def create_swap(self, drives):
        for drive in drives:
            self.storage.createDevice(self.storage.newPartition(fmt_type="swap",
                                                size=self.instClass.swap_size,
                                                parents=drive,
                                                weight=1000))
        log.info("Created swap instance on %s" % ', '.join(d.name for d in drives))

    @property
    def vstorage_slash_max_size(self):
        return max(self.instClass.max_slash_size, self.instClass.slash_size +
            len([d for d in self.storage.disks if d.use_csdata]) * Size("1536MiB"))

    def device_func(self, device):
        device_func = self.storage.newLV
        if device.isDisk:
            device_func = self.storage.newPartition
        return device_func

    def create_slash(self, devices, maxsize, vg = False):
        if vg:
            self.storage.createDevice(self.storage.newLV(fmt_type="ext4",
                                                size=self.instClass.slash_size,
                                                mountpoint="/",
                                                parents=[self.vg],
                                                grow=True,
                                                maxsize=maxsize))
            log.info("Created / instance on VG %s" % self.vg.name)
            return
        md = MDdev(self.storage, devices, "ext4", self.instClass,
            size=self.instClass.slash_size, mountpoint="/", grow=True, maxsize=maxsize)
        md.create_members()
        self.slash_mddevs.append(md)
        log.info("Created / instance on %s" % ', '.join([ d.name for d in devices ]))

    def create_lvm(self, pv_drives):
        # No drives - no volume
        if not pv_drives:
            return None

        # Prepare PVs
        pv_parts = []
        for disk in pv_drives:
            part = self.storage.newPartition(fmt_type="lvmpv",
                                                grow=True,
                                                parents=[disk])
            self.storage.createDevice(part)
            pv_parts.append(part)
            log.info("Created pv instance on %s" % disk.name)

        # Create LVM VG volume
        vg = self.storage.newVG(parents=pv_parts)
        self.storage.createDevice(vg)
        return vg

    def create_vz(self, drives, grow=True):
        if None in drives:
            return

        self.storage.createDevice(self.storage.newLV(fmt_type="ext4vz",
                                                size=self.instClass.min_vz_size,
                                                mountpoint="/vz",
                                                grow=grow,
                                                parents=drives))
        log.info("Created /vz instance on %s" % ', '.join(d.name for d in drives))

    @property
    def get_slash_drives(self):
        return [ d for d in self.storage.disks if d.use_slash ]

    @property
    def get_vz_drives(self):
        return [ d for d in self.storage.disks if d.use_vzdata ]

    @property
    def get_ssd_drives(self):
        return [ d for d in self.storage.disks if d.use_ssd ]

    def create_ssd(self, device):
        self.ssdnum += 1
        self.storage.createDevice(self.storage.newPartition(fmt_type="ext4",
                                                mountpoint="/mnt/ssd%i" % self.ssdnum,
                                                size=self.instClass.min_ssd_size,
                                                grow=True,
                                                parents=[device]))
        log.info("Created SSD cache instance on %s" % device.name)

    def create_ssds(self, drives):
        [ self.create_ssd(d) for d in drives ]

    @property
    def get_cs_disks(self):
        return [ d for d in self.storage.disks if d.use_csdata and not d.use_slash ]

    def creats_cs(self, disk):
        self.csnum += 1
        self.storage.createDevice(self.device_func(disk)(fmt_type="ext4cs",
                                                mountpoint="/vstorage/%s-cs%i" %
                                                (self.storage.vstorage_cs, self.csnum),
                                                size=(self.instClass.min_cs_size - Size("1MiB")),
                                                grow=True,
                                                parents=[disk]))
        log.info("Created CS instance on %s" % disk.name)

    def create_vstorage_cs(self, devices):
        [ self.creats_cs(d) for d in devices ]

    @property
    def slash_together_cs(self):
        return [ d for d in self.get_slash_drives if d.use_csdata ]

    @property
    def slash_together_vz(self):
        return [ d for d in self.get_slash_drives if d.use_vzdata ]

    @property
    def is_multislash(self):
        return len(self.get_slash_drives) > 1

    def create_not_vstorage_and_not_together(self):
        self.prepare_boot(self.get_boot_disks)
        self.create_swap(self.get_slash_drives)
        self.vg = self.create_lvm(self.get_vz_drives)
        self.create_slash(self.get_slash_drives, 0) # 0 means maximum available size
        self.create_vz([self.vg])
        self.create_ssds(self.get_ssd_drives)

    def create_not_vstorage_and_together(self):
        self.prepare_boot(self.get_boot_disks)
        self.create_swap(self.get_slash_drives)
        self.vg = self.create_lvm(self.get_vz_drives)
        if self.is_multislash:
            self.create_slash(self.get_slash_drives, self.instClass.max_slash_size)
        else:
            self.create_slash([], self.instClass.max_slash_size, vg=True)
        self.create_vz([self.vg])
        self.create_ssds(self.get_ssd_drives)

    def create_vstorage_and_not_together(self):
        self.prepare_boot(self.get_boot_disks)
        self.create_swap(self.get_slash_drives)
        self.vg = self.create_lvm(self.get_slash_drives)
        if self.is_multislash:
            self.create_slash(self.get_slash_drives, self.vstorage_slash_max_size)
        else:
            self.create_slash([], self.vstorage_slash_max_size, vg=True)
        self.create_vz([self.vg])
        self.create_ssds(self.get_ssd_drives)
        self.create_vstorage_cs(self.get_cs_disks)

    def create_vstorage_and_together(self):
        self.prepare_boot(self.get_boot_disks)
        self.create_swap(self.get_slash_drives)
        self.vg = self.create_lvm(self.get_slash_drives)
        if self.is_multislash:
            self.create_slash(self.get_slash_drives, self.vstorage_slash_max_size)
        else:
            self.create_slash([], self.vstorage_slash_max_size, vg=True)
        self.create_vz([self.vg], grow=False)
        self.create_ssds(self.get_ssd_drives)
        self.create_vstorage_cs(self.get_cs_disks)
        # And create vstorage CS on slash PV
        self.create_vstorage_cs([self.vg])

    def validate_selection(self):
        errormess = ""

        if not self.get_slash_drives:
            errormess = _("At least one drive for / need to be selected.\n")
        if self.instClass.is_vz and not self.get_vz_drives:
            errormess += _("At least one drive for /vz need to be selected.\n")

        return errormess

    def log_preselect(self):
        for disk in self.storage.disks:
            log.info("Flags for %s" % disk.name)
            if disk.cap_slash:
                log.info("%s capable for slash" % disk.name)
            if disk.use_slash:
                log.info("%s selected for slash" % disk.name)
            if disk.cap_vzdata:
                log.info("%s capable for VZ" % disk.name)
            if disk.use_vzdata:
                log.info("%s selected for VZ" % disk.name)
            if disk.cap_csdata:
                log.info("%s capable for CS" % disk.name)
            if disk.use_csdata:
                log.info("%s selected for CS" % disk.name)
            if disk.is_ssd:
                log.info("%s is SSD" % disk.name)
            if disk.use_ssd:
                log.info("%s selected for SSD" % disk.name)

    def execute(self):
        if not self.storage.can_autopart:
            return

        self.log_preselect()

        # Reset drive numbers
        self.csnum = 0
        self.ssdnum = 0

        # Check selection
        errormess = self.validate_selection()
        if errormess:
            raise PartitioningError(_("autopart failed:\n%s") % errormess)

        self.storage.doAutoPart = True

        clear_all_partitions(self.storage)

        # Check for OpenVZ - disable vstorage case
        if self.instClass.is_free:
            self.storage.vstorage_cs = ""

        # Will process 4 scenarios
        # Not together means that /vz placed on separate drive from /
        if not self.storage.vstorage_cs and not self.slash_together_vz:
            self.create_not_vstorage_and_not_together()
        # Together means that /vz placed on the same drive with /
        elif not self.storage.vstorage_cs and self.slash_together_vz:
            self.create_not_vstorage_and_together()
        elif self.storage.vstorage_cs and not self.slash_together_cs:
            self.create_vstorage_and_not_together()
        elif self.storage.vstorage_cs and self.slash_together_cs:
            self.create_vstorage_and_together()

        # Create members before partitions fixage to do proper size calculations
        [ md.create_members() for md in self.mddevs ]

        doPartitioning(self.storage)

        [ md.create_md() for md in self.mddevs + self.slash_mddevs ]

        growLVM(self.storage)

        self.storage.setUpBootLoader()

        # only newly added swaps should appear in the fstab
        self.storage.setFstabSwaps((dev for dev in self.storage.swaps if not dev.format.exists))

        report = storage_utils.storage_checker.check(self.storage)
        report.log(log)

        if report.failure:
            raise PartitioningError(_("autopart failed: \n") + "\n".join(report.all_errors))

def do_try(fn, *args, **kwargs):
    try:
        return fn(*args, **kwargs)
    except:
        pass

def clear_all_partitions(storage):
    storage.config.clearPartType = CLEARPART_TYPE_ALL
    storage.config.clearPartDisks = [ d.name for d in getDisks(storage.devicetree) ]

    # GPT only!
    default_disklabel_type = "gpt"
    if not platform.setDefaultDiskLabelType(default_disklabel_type):
        raise PartitioningError(_("Failed to set default disktype to %s\n") %
            default_disklabel_type)

    # TODO Fill drivers to clear in GUI
    storage.clearPartitions()

def rpm_import(src):
    """
    Import the RPM GPG key to the installed system.
    Input parameter is the full path to a key.
    """

    log.info("Import the RPM key: %s" % src)

    rc = execWithRedirect("/bin/rpm", [ "--import", src,
                          "--dbpath", getSysroot() + "/var/lib/rpm"])

    if rc:
        log.error("Failed to import the RPM key %s, error(%i)" % (src, rc))

def calculate_ver(m):
    version = 0
    if m and m.group(1) and m.group(2) and m.group(3) and m.group(4):
        version = int(m.group(1)) * pow(100,7) + int(m.group(2)) * pow(100,6) \
            + int(m.group(3)) * pow(100,4) + int(m.group(4))
    return version

def getver_str(ver, rel=True):
    major = ver / pow(100,7)
    middle = (ver - major * pow(100,7)) / pow(100,6)
    minor = (ver - major * pow(100,7) - middle * pow(100,6)) / pow(100,4)
    release = ver - major * pow(100,7) - middle * pow(100,6) - minor * pow(100,4)
    if rel:
        return "%i.%i.%i (%i)" % (major, middle, minor, release)
    return "%i.%i.%i" % (major, middle, minor)

def check_va_repo(url, instClass, repo_suffix):
    if type(url) == list:
        try:
            url = url[0]
        except:
            return ""
    if not url:
        return ""
    baseurl = url + "/" + repo_suffix
    testurl = baseurl + "/repodata/repomd.xml"
    try:
        urlgrabber.urlopen(testurl)
    except urlgrabber.grabber.URLGrabError as e:
        log.error("Failed to get baserepo url %s" % testurl)
        return ""
    return baseurl

def check_va_mirrorlist(mirrorlist, instClass, repo_suffix):
    if not mirrorlist:
        return ""
    try:
        f = urlgrabber.urlopen(mirrorlist)
        for m in [ m for m in f.readlines() if m and not m.startswith('#') ]:
            baseurl = check_va_repo(m.rstrip('\n'), instClass, repo_suffix)
            if baseurl:
                return baseurl
    except urlgrabber.grabber.URLGrabError as e:
        log.error("Failed to get mirrorlist %s" % mirrorlist)
        return ""
    return ""

def get_base_repos(payload, instClass):
    # Attach base repo
    if not isinstance(payload, YumPayload):
        log.error("Failed to setup base repos: not yum payload")
        return

    for repo in [ r for r in payload._yum.repos.repos.values() if r.id == BASE_REPO_NAME ]:
        if repo.baseurl and repo.baseurl[0]:
            instClass.base_repos.append(repo.baseurl[0])

def get_va_repos(payload, instClass):
# TODO - not work w/o kickstart or va-agent in list
#    if ksdata.registervaagent.enabled or "va-agent" in ksdata.packages.groupList:

    # Attach va-agent repo from base repo
    if not isinstance(payload, YumPayload):
        log.error("Failed to setup Agent and MN repos: not yum payload")
        return

    for repo in [ r for r in payload._yum.repos.repos.values() if r.id == BASE_REPO_NAME ]:
        # Try mirrorlists first
        if repo.mirrorlist:
            instClass.base_mlists.append(repo.mirrorlist)
        brepo = check_va_mirrorlist(repo.mirrorlist, instClass, instClass.va_agent_repo)
        if brepo:
            instClass.base_va_agent_repos.append(brepo)
        brepo = check_va_mirrorlist(repo.mirrorlist, instClass, instClass.va_mn_repo)
        if brepo:
            instClass.base_va_mn_repos.append(brepo)
        if instClass.base_va_mn_repos and instClass.base_va_agent_repos:
            break
        # And then repositories
        if repo.baseurl and repo.baseurl[0]:
            instClass.base_repos.append(repo.baseurl[0])
        brepo = check_va_repo(repo.baseurl, instClass, instClass.va_agent_repo)
        if brepo:
            instClass.base_va_agent_repos.append(brepo)
        brepo = check_va_repo(repo.baseurl, instClass, instClass.va_mn_repo)
        if brepo:
            instClass.base_va_mn_repos.append(brepo)
        if instClass.base_va_agent_repos or instClass.base_va_mn_repos:
            break

    # base repos failed
    if not instClass.base_va_agent_repos and not instClass.base_va_agent_mlists:
        # Use net
        instClass.base_va_agent_mlists = instClass.net_va_agent_mirrors
    if not instClass.base_va_mn_repos and not instClass.base_va_mn_mlists:
        # Use net
        instClass.base_va_mn_mlists = instClass.net_va_mn_mirrors

    if not instClass.base_repos and not instClass.base_mlists:
        # Use net
        instClass.base_mlists = instClass.net_base_mirrors

    for r in instClass.base_va_agent_repos + instClass.base_va_agent_mlists + \
        instClass.base_va_mn_repos + instClass.base_va_mn_mlists + \
        instClass.base_repos + instClass.base_mlists:
        log.info("Repository configured: %s" % r)

def install_va_agent(payload, ksdata, instClass):
    # Should not happens, but
    if not instClass.base_va_agent_repos and not instClass.base_va_agent_mlists:
        log.error("Failed to get agent base repos and mirrorlists")
        return False

    # TODO: add proxy info if any
    for repo_num in range(len(instClass.base_va_agent_repos)):
        payload.addRepo(ksdata.RepoData(name="installer-va-agent-%i" % repo_num,
                                 baseurl=instClass.base_va_agent_repos[repo_num],
                                 enabled=True))
        log.info("Adding va-agent repo %s", instClass.base_va_agent_repos[repo_num])

    for mlist_num in range(len(instClass.base_va_agent_mlists)):
        payload.addRepo(ksdata.RepoData(name="installer-va-agent-%i" % mlist_num,
                                 mirrorlist=instClass.base_va_agent_mlists[mlist_num],
                                 enabled=True))
        log.info("Adding va-agent mirrorlist %s", instClass.base_va_agent_mlists[mlist_num])

    payload.updateBaseRepo()
    payload.selectGroup("va-agent")
    payload.selectPackage(instClass.agent_release)
    # TODO: check for vstorage-ui group installed
    payload.selectPackage("va-agent-storage")

    return True

def install_vstorage_ui(payload):
    payload.selectGroup("vstorage")
    payload.selectGroup("vstorage-ui-agent")

def add_trusted_interface(ksdata):
    # Extract IP address specified for the storage roles
    vstorage = getattr(ksdata, 'vstorage')
    if not vstorage:
        return
    ipaddr = getattr(vstorage, 'ipaddr')
    if not ipaddr and len(vstorage.mds_create) == 2:
        ipaddr = vstorage.mds_create[1]
    if not ipaddr and len(vstorage.mds_join) == 2:
        ipaddr = vstorage.mds_join[1]
    if not ipaddr:
        return

    # Identify network device by this IP address
    device = None
    for devname in nm_activated_devices():
        if filter(lambda m: m == ipaddr, get_full_ip_addresses(devname, version=4, with_prefix=False) + get_full_ip_addresses(devname, version=6, with_prefix=False)):
            device = devname
            break
    if not device:
        log.error('Could not find network device by its IP %s.' % (ipaddr))
        return

    # Add this device (and the bridge it is a part of, if any) into the 'trusted' zone of firewall
    ifname = device
    while ifname:
        zoned = False
        bridge = None
        filename = os.path.join(getSysroot(), 'etc/sysconfig/network-scripts/ifcfg-%s' % (ifname))
        if not os.path.isfile(filename):
            log.error('Failed to add network device %s to trusted: ifcfg script is missing.' % (ifname))
            break
        lines = open(filename, 'r').readlines()
        for (i, ln) in enumerate(lines):
            ln = ln.rstrip('\n')
            m = re.match(r'^BRIDGE=[\'\"]?([^\'\"]+)[\'\"]?$', ln, re.I)
            if m:
                bridge = m.group(1)
            m = re.match(r'^ZONE=[\'\"]?([^\'\"]+)[\'\"]?$', ln, re.I)
            if m:
                lines[i] = 'ZONE=trusted\n'
                zoned = True
        if not zoned:
            lines.append('ZONE=trusted\n')
        open(filename, 'w').writelines(lines)
        log.info('Added device %s (IP: %s) to the trusted zone.' % (ifname, ipaddr))
        if bridge != ifname:
            ifname = bridge
        else:
            ifname = None

def preInstall(storage, payload, ksdata, instClass):
    # For UI agent
    for unit in ("vstorage-ui-agent", "nginx"):
        stop_service(unit)
    if instClass.is_free:
        return
    if not instClass.is_vz:
        get_base_repos(payload, instClass)
        return
    if [ c for c in ksdata.container.containerList if c.special ]:
        payload.selectPackage("%s-ez" % instClass.defostemplate)
    # Fetch repos
    get_va_repos(payload, instClass)
    if ksdata.registervaagent.enabled or ksdata.registervaagent.install_pkgs:
        install_va_agent(payload, ksdata, instClass)
    if ksdata.registervstorageui.enabled:
        install_vstorage_ui(payload)
    # Vstorage - select group if clustername given
    if ksdata.vstorage.clustername:
        payload.selectGroup("vstorage")
        if ksdata.vstorage.client and ksdata.vstorage.shaman:
            payload.selectGroup("high-availability")
# Temporary disabled: need to think whats to do in such case
#        if not ksdata.vstorage.client:
#            map(lambda x: payload.deselectGroup(x), ["vz", "ps", "templates"])

def postInstall(storage, ksdata, instClass):
    if instClass.create_bridges:
        do_try(execInSysroot, "/usr/libexec/create_bridges.py", [])

    if instClass.fix_slave_cfgs:
        do_try(execInSysroot, "/usr/libexec/fix-slave-cfgs.py", [])

    if instClass.fix_ifcfgs:
        do_try(execInSysroot, "/usr/libexec/fix-devices.py", [])

    if instClass.is_vz:
        do_try(execInSysroot, "/usr/libexec/vzfirewall.py", ['add'])
        add_trusted_interface(ksdata)
    else:
        do_try(execInSysroot, "/usr/bin/systemctl", ['disable', 'firewalld'])

    if instClass.mask_nm:
        for s in ("NetworkManager-dispatcher.service", "NetworkManager.service", "NetworkManager-wait-online.service", "firewalld.service"):
            do_try(execWithRedirect, "systemctl", ["mask", s], root=getSysroot())

    process_opensm(cmds=("enable",), root=getSysroot())

    product_name = ""
    product_releasef = instClass.get_release_f
    repover = []

    # Detect version ISO
    repover.append(calculate_ver(re.match("^.*(\d+)\.(\d+)\.(\d+) \((\d+)\).*", productVersion, re.I)))

    # Correct release file
    try:
        with open(getSysroot() + product_releasef, "r+") as f:
            data = f.read()
            product_name = re.sub(" [0-9]*\.[0-9]*\.[0-9]* \([a-zA-Z0-9\.]*\).*", "", data).rstrip()
            repover.append(calculate_ver(re.match("^.*(\d+)\.(\d+)\.(\d+) \((\d+)\).*", data, re.I)))
            f.seek(0)
            f.write("%s %s\n" % (product_name, getver_str(sorted(repover, reverse=True)[0])))
            f.truncate()
            f.close()
    except IOError as e:
        log.error("Failed to correct %s: %s" % (product_releasef, e))

    try:
        with open( getSysroot() + "/etc/vz/ploop.conf", "w") as f:
            f.write("USE_KAIO_FOR_EXT4=yes\n")
    except IOError as e:
        log.error("Failed to write ploop.conf: %s" % (product_releasef, e))

    src = getSysroot() + "/etc/pki/rpm-gpg"
    [ rpm_import(os.path.join(src, i)) for i in os.listdir(src) ]

def create_vstorage_input(pw):
    pipe = os.pipe()
    os.write(pipe[1], pw + '\n')
    os.close(pipe[1])
    return pipe[0]

# Checks that the file is present or, if not, downloads it to the specified location.
# Returns the file location.
def bring_file(instClass, source, destination, force_copy = True):
    # Check for base repos
    if not instClass.base_repos:
        log.error("No base repositories configured!")
        return False

    # Check if requested file is already there
    if os.path.exists(destination):
        log.info("Destination file %s already exists." % destination)
        return destination

    dest_dir = os.path.dirname(destination)
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)

    local_paths = []
    remote_paths = []
    for brepo in instClass.base_repos:
        sourcepath = "%s/%s" % (brepo, source)
        prefix = 'file://'
        if sourcepath.startswith(prefix):
            local_paths.append(sourcepath[len(prefix):])
        else:
            remote_paths.append(sourcepath)

    # First, try to search locally
    for sourcepath in local_paths:
        if os.path.isfile(sourcepath):
            if force_copy:
                shutil.copy(sourcepath, destination)
                log.info('Copied file %s to %s' % (sourcepath, destination))
                return destination
            else:
                log.info('Found requested file at %s, not requested to copy' % (sourcepath))
                return sourcepath

    # Failed, try remote
    for sourcepath in remote_paths:
        try:
            urlgrabber.urlgrab(sourcepath, filename=destination, copy_local=1)
            log.info('Downloaded file from %s to %s' % (sourcepath, destination))
            return destination
        except (IOError, urlgrabber.grabber.URLGrabError) as e:
            log.error('Failed to download URL %s to %s: %s' % (sourcepath, destination, e))
            do_try(os.unlink, destination)
            continue

    log.error('File %s was not found!' % (source))
    return None

def check_vstorage_avail():
    if not os.path.exists(vstorage_bin):
        return False
    return True

def process_vstorage_auth_dirs(cluster_name):
    keyfile_dir = "/etc/vstorage/clusters/%s" % cluster_name
    keyfile = os.path.join(keyfile_dir, "auth_digest.key")

    if not os.path.exists(keyfile_dir):
        os.makedirs(keyfile_dir)

    if os.path.exists(keyfile):
        os.unlink(keyfile)

def vstorage_auth_node(cluster_name, vstorage_clusterpw, reg_timeout, stage2=False):
    if stage2 and not check_vstorage_avail():
        log.error("%s absent in 2nd stage" % vstorage_bin)
        return 255

    if stage2:
        rootpath = "/"
        process_vstorage_auth_dirs(cluster_name)
    else:
        rootpath = getSysroot()

    while (reg_timeout):
        input_data = create_vstorage_input(vstorage_clusterpw)
        auth_start_time = time.time()
        rc = execWithRedirect("/usr/bin/vstorage", [ "-vvvvvvvv", "-t", "30", "-c", cluster_name,
            "auth-node", "-P"], root=rootpath, stdin=input_data)
        overal_auth_time = time.time() - auth_start_time
        os.close(input_data)
        log.debug("vstorage auth-node exit code is %i, "
            "working time: %.2f sec." % (rc, overal_auth_time))

        if rc == 0 or rc == 253 or reg_timeout == 1:    #code 253 - wrong password.
            break

        if overal_auth_time > 60:
            log.warning("Time exceeded execution of command "
                "vstorage auth-node: %.2f sec." % overal_auth_time)
        else:
            time.sleep(60 - overal_auth_time)
        reg_timeout -= 1

    if rc == 0:
        log.info("The server has been successfully authenticated.")
    elif rc == 253:
        log.error("Failed to authenticate the server in the cluster: wrong password")
    else:
        log.error("Authentication failed: Could not connect to the cluster! " \
            "see vstorage error(s) in anaconda program log")

    return rc

def checkVstorageLicense(clustername, license):
    if not check_vstorage_avail():
        log.error("%s absent in 2nd stage" % vstorage_bin)
        return ""

    buf = execWithCapture(vstorage_bin, [ "view-license", "-p", license ])

    log.info("Checking entered %s key %s" % (clustername, license))

    if "status=\"VALID\"" in buf or "keyserver_host=" in buf:
        return license

    log.error("Entered key %s is not activation %s key or valid key.:\n%s" % \
            (license, clustername, buf))
    return ""

def get_full_ip_addresses(devname, version=4, with_prefix=False):
    try:
        addrs = []
        ip_config = nm_device_ip_config(devname, version)
        if not ip_config:
            return []
        for addr in ip_config[0]:
            if not addr:
                continue
            if with_prefix:
                addrs.append("%s/%i" % (addr[0], addr[1]))
                continue
            addrs.append(addr[0])
        return addrs
    except (dbus.DBusException, ValueError) as e:
        log.warning("Got an exception trying to get the ip addr "
                    "of %s: %s", devname, e)

    return []

def get_nonlocal_ip_list(ipv4_only = True, with_prefix=False):
    ip_list = []
    ipv4_addresses = []
    ipv6_addresses = []
    for devname in nm_activated_devices():
        ipv4_addresses += get_full_ip_addresses(devname, version=4, with_prefix=with_prefix)
        if not ipv4_only:
            ipv6_addresses += get_full_ip_addresses(devname, version=6, with_prefix=with_prefix)

    for ip in ipv4_addresses + ipv6_addresses:
        try:
            if IPNetwork(ip).is_loopback() or IPNetwork(ip).is_link_local():
                continue
        except:
            # Looks like bad address
            continue
        ip_list.append(ip)

    if len(ip_list) == 0:
        log.warning("No network interface card on this server has "
                "a valid IP address assigned.\nMake sure the "
                "network for this server is configured properly.")

    return ip_list

# Equivalent to standart execWithRedirect, but close stdout after execution
def execWithRedirectCl(*argv, **kwargs):
    rc = execWithRedirect(*argv, **kwargs)
    try:
        kwargs["stdout"].close()
    except:
        pass
    return rc

def inside_vm():
    """Check if the installer is running inside a VM."""

    return hvt_checker.CPU().has_feature("hypervisor")

def vt_supported():
    """
    Check that basic virtualization is enabled and the CPU has full support
    of required features.
    """

    cpu = hvt_checker.CPU.get()
    if cpu is None:
        return False

    return cpu.vt_enabled() and cpu.full_featured() and cpu.kvm_loaded()

# Check that drive fast enough to be SSD
def is_ssd(disk):
    # HACK HACK HACK - provide ssd_hack=sda,sdb,sdg for SSDs
    try:
        f = open("/proc/cmdline", "r")
        for param in f.read().split(" "):
            if not param.startswith("ssd_hack="):
                continue
            for ssd in re.sub("^ssd_hack=", "", param).split(","):
                if ssd.rstrip('\n') == disk.name:
                    return True
        f.close()
    except:
        pass

    if inside_vm():
        return False

    ssd_iops_barrier = 2000

    # Use precached values
    if disk.iops:
        log.info("IOPS for %s already measured: %i\n" % (disk.name, disk.iops))
        if disk.iops > ssd_iops_barrier:
            return True
        return False

    devicename = "/dev/%s" % disk.name

    ioping_out = execWithCapture("/usr/bin/ioping",
        [ "-D",  "-B",  "-q",  "-i",  "0", "-c", "10k", "-w", "1s", devicename ],
        filter_stderr=True)
    ioping = re.match("^[0-9]+ [0-9]+ (\d+) .*$", ioping_out, re.I)
    if ioping and ioping.group(1):
        log.info("Measured IOPS for %s: %s" % (devicename, ioping.group(1)))
        disk.iops = int(ioping.group(1))
        if int(ioping.group(1)) > ssd_iops_barrier:
            log.info("Detected SSD drive: %s" % devicename)
            return True
    else:
        log.info("Failed to measure IOPS for %s" % devicename)

    return False

def minimal_hardware(disks, join_cs):
    if join_cs:
        return not bool([ d for d in disks if d.cap_csdata ])
    return not bool([ d for d in disks if d.cap_vzdata ])

def in_cmdline(cparam):
    result = False
    try:
        f = open("/proc/cmdline", "r")
        for param in f.read().split(" "):
            if param.rstrip('\n') == cparam:
                result = True
                break
        f.close()
    except:
        pass

    return result

def start_containers(ksdata, special, addr):
    containers_to_start = []

    for c in [ c for c in ksdata.container.containerList if c.special == special ]:
        for ip in c.ip:
            if ip == addr:
                containers_to_start.append(c)
        if c.hostname == addr:
            containers_to_start.append(c)

    for c in containers_to_start:
        if c.bind(c.to_bind) and \
                c.setup_vz("start") and \
                c.start_container():
            log.info("Started %s container: %s" % (special, c.container_id))
        else:
            log.error("Failed to start %s container: %s" % (special, c.container_id))
            containers_to_start.remove(c)

    return containers_to_start

def stop_containers(containers_to_start):
    for c in containers_to_start:
        c.cleanup()

def vstorage_init_runned():
    return threadMgr.get(THREAD_VSTORAGE_WAIT_STORAGE_INIT)

def start_wait_vstorage_init(handler):
    if not vstorage_init_runned():
        threadMgr.add(AnacondaThread(name=THREAD_VSTORAGE_WAIT_STORAGE_INIT, target=handler))

def payload_err(e):
    log.error(e)
    exn = PayloadInstallError(e)
    if errorHandler.cb(exn) == ERROR_RAISE:
        progressQ.send_quit(1)
        ipmi_report(IPMI_ABORTED)
        sys.exit(1)

def copy_liveos(payload, instClass):
    if not instClass.is_squashroot:
        return
    logfile = "/tmp/copy_liveos.log"
    try:
        os.unlink("/etc/fstab")
    except:
        pass
    tmpdir = "/tmp/liveroot"
    os.mkdir(tmpdir)
    rc = execWithRedirectCl("/bin/mount", ["-o", "bind", "/", tmpdir],
        stdout = open(logfile, 'a+'))
    if rc:
        payload_err("Failed to bind-mount /")

    # preserve: permissions, owners, groups, ACL's, xattrs, times,
    #           symlinks, hardlinks
    # go recursively, include devices and special files, don't cross
    # file system boundaries
    cmd = "/usr/bin/rsync"
    try:
        rc = execWithRedirectCl(cmd, ["-pogAXtlHrDx",
            "--exclude", "/dev/", "--exclude", "/proc/", "--exclude",
            "/sys/", "--exclude", "/run/", "--exclude", "/boot/*rescue*",
            tmpdir+"/", getSysroot()], stdout = open(logfile, 'a+'))
    except (OSError, RuntimeError) as e:
        payload_err(str(e))

    execWithRedirectCl("/bin/umount", [tmpdir], stdout = open(logfile, 'a+'))

    rc = execWithRedirectCl("/usr/libexec/unsquash.sh", [], root=getSysroot(),
        stdout = open(logfile, 'a+'))
    if rc:
        payload_err("Failed to unsquashfs image: %i" % rc)

def wait_network_devices(class_name, msg):
    hubQ.send_not_ready(class_name)
    if msg:
        hubQ.send_message(class_name, msg)

    # wait for network
    if threadMgr.get(THREAD_WAIT_FOR_CONNECTING_NM):
        threadMgr.wait(THREAD_WAIT_FOR_CONNECTING_NM)

    # wait for network devices
    while network_devices_available() and not network_devices_active():
        # Installation started - exit, skip network wait
        if threadMgr.get(THREAD_INSTALL):
            return
        time.sleep(1)

    hubQ.send_ready(class_name, True)

def start_wait_network_devices(class_name, thread_name, msg):
    if network_devices_available() and not network_devices_active():
        if not threadMgr.get(thread_name):
            threadMgr.add(AnacondaThread(name=thread_name, target=wait_network_devices, args=(class_name, msg)))
        return False
    return True

def network_devices_active():
    return get_nonlocal_ip_list(False)

def network_devices_available():
    return nm_devices()

def get_progress_hub():
    if not expert_mode() and os.path.exists("/etc/hci-release"):
        return "SummaryHub"

    return "ProgressHub"

def runPreExitScripts(storage, ksdata, instClass):
    for fn in glob.glob("/usr/share/anaconda/pre-exit-scripts/*"):
        rc = execWithRedirectCl(fn, [], stdout = open("/dev/tty7", 'a+'))
        if rc:
            log.error("Failed to execute %s: %i" % (fn, rc))

def anaconda_mode():
    try:
        return os.readlink("/etc/systemd/system/default.target").endswith("anaconda.target")
    except:
        return True
