from pyanaconda.i18n import _
import pykickstart.commands as commands
from pykickstart.errors import formatErrorMsg, KickstartError, KickstartValueError
from pyanaconda.iutil import execInSysroot, execWithRedirect, getSysroot, execWithCapture
from pyanaconda import network
from pyanaconda.flags import flags
from pyanaconda.virtuozzo_c import vstorage_auth_node, create_vstorage_input, \
    get_nonlocal_ip_list, execWithRedirectCl, checkVstorageLicense, DoVZAutopart, \
    start_containers, stop_containers, vt_supported, create_vstorage_input, \
    bring_file
from pyanaconda.packaging.yumpayload import YumPayload
from netaddr import IPNetwork, IPAddress
from copy import copy
import shutil
from logging import getLogger
log = getLogger("anaconda")
import os
import re
import warnings
import time
import uuid
import tempfile
import signal
import socket
import random
from blivet.size import Size

class Vstorage(commands.vstorage.VZ7_Vstorage):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.vstorage.VZ7_Vstorage.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.mds_port = 2510
        self.started_services = []
        self.vstorage_bin = "/usr/bin/vstorage"
        self.common_opts = ["-vvvvvvvv", "-t", "90", "-c" ]
        self.logfile = "/tmp/vstorage.log"
        self.shaman_bin = "/usr/sbin/shaman"
        self.snmp_mib_tree = ".1.3.6.1.4.1.26171"
        self.client_dir = ""
        self.mds_min_size = Size("10GiB")
        # Save 20% of data on SSDs for reserve, backported from PCS6
        self.ssd_reserved_coeff = 0.8
        self.default_vstorage_path = "/vstorage"
        self.vstorage_service = "/usr/libexec/vstorage/vstorage-service"

    def stop_service(self, service):
        rc = execInSysroot("killall", [service])
        if rc:
            log.info("Failed to stop %s service: %i . Looks like already stopped", service, rc)
            return False
        return True

    def start_service(self, service, args):
        if not os.path.exists(getSysroot() + self.vstorage_service):
            log.error("VStorage service binary %s does not exist", self.vstorage_service)
            return False
        rc = execInSysroot(self.vstorage_service, ["start", self.clustername, service] + args)
        if rc:
            log.error("Failed to start %s service: %i", service, rc)
            return False
        self.started_services.append(service)
        return True

    def stop_services(self):
        for service in self.started_services:
            self.stop_service(service)

    def determine_subnet(self, ip):
        """Determine subnet address in CIDR notation.
           Return subnet address in string format or None"""
        net_info = execWithCapture("ip", ["addr", "show"], filter_stderr=True)

        data = re.search('%s[^\s]*' % ip, net_info)
        if not data:
            return None

        return str(IPNetwork(data.group()).cidr)

    def get_vstorage_client_ip(self):
        """Retrieve ip address of client from 'vstorage stat' output.
           Here we assume that this address can be used to contact cluster.
           Return ip address in string format or None"""
        stat = execWithCapture("vstorage", ["-c", self.clustername, "stat", "-X"],
            filter_stderr=True, root=getSysroot())

        cluster_ips = set(re.findall('<host>([^\<host>]*)', stat))
        node_ips = set(get_nonlocal_ip_list())

        client_ip = node_ips & cluster_ips
        if not client_ip:
            return None

        return client_ip.pop()

    def configure_snmpd(self):
        ip = self.get_vstorage_client_ip()
        if not ip:
            log.error("Failed to determine client ip address")
            return

        subnet = self.determine_subnet(ip)
        if not subnet:
            log.error("Failed to retrieve cluster subnet address")
            return

        try:
            f = open(getSysroot() + "/etc/snmp/snmpd.local.conf", "a")

            try:
                sample = "rwcommunity" + "\tpdrs" + "\t%s" + "\t%s"

                # grant read/write access for localhost
                print >> f, sample % ("127.0.0.1", self.snmp_mib_tree)
                # grant read/write access for cluster subnet
                print >> f, sample % (subnet, self.snmp_mib_tree)

            finally:
                f.close()

        except IOError:
            log.error("Failed to edit configuration file of the SNMP daemon")

    def prepare_ssd_sizes(self, ssds, clients):
        space_vals = []
        # Sort - more diskspace - first
        ssds = sorted(ssds, key=lambda k: k[1], reverse = True)

        # Variant 1 - less or eq clients then ssds: sort and use max
        if clients <= len(ssds):
            # Set equal sizes for all ssds
            for mp in sorted(set([ d[0] for d in ssds ])):
                sizes = [ d for d in ssds if d[0] == mp ]
                splitted = len(sizes)
                allsize = sum([ d[1] for d in sizes ])
                for ssd in sizes:
                    ssd[1] = allsize/splitted
            for i in range(clients):
                space_vals.append((ssds[i][0], ssds[i][1]))
            return space_vals

        # Variant2 - clients more than SSDs, so need to split smth
        # Will split bigger, then resort and requrse and so one
        ssds[0][1] /= 2
        ssds.append(copy(ssds[0]))

        return self.prepare_ssd_sizes(ssds, clients)

    def get_ssd_caches(self, storage):
        ssd_cache_parts = []
        for mp in sorted([ mp for mp in storage.mountpoints if
            mp.startswith("/mnt/ssd") ]):
            # Calculate size
            try:
                statfs = os.statvfs(getSysroot() + mp)
                size = int(statfs.f_bsize * statfs.f_bavail / Size("1MiB"))
                ssd_cache_parts.append([ mp, int(size * self.ssd_reserved_coeff) ])
                log.info("Detected SSD cache mounted at %s size as %s MiB" % (mp, size))
            except Exception, e:
                log.error("Can't detect available ssd cache size for %s\nError: %s" % (mp, e))

        if not ssd_cache_parts:
            return []

        clients = 0
        if self.need_mds_create():
            # Reserve size on 1st drive for MDS
            ssd_cache_parts[0][1] -= self.mds_min_size/Size("1MiB")
        if self.cs_join and self.cs_ssd:
            clients += len(self.vstorage_parts(storage))
        if self.client and self.client_ssd:
            clients += 1

        return self.prepare_ssd_sizes(ssd_cache_parts, clients)

    def vstorage_parts(self, storage):
        return [ p for p in storage.partitions + storage.lvs +
            storage.mdarrays if p.format.type == "ext4cs" ]

    def execute(self, storage, ksdata, instClass):
        if not self.clustername or instClass.is_free:
            return True

        if not os.path.exists(getSysroot() + self.vstorage_bin):
            log.error("%s absent in installed system" % self.vstorage_bin)
            return False

        if not self.need_mds_create():
            vstorage_auth_node(self.clustername, self.clusterpw, self.reg_timeout + 1)

        # Calculate sizes for SSD caches
        ssd_caches = self.get_ssd_caches(storage)

        # Check for main directory
        if not os.path.exists(getSysroot() + self.default_vstorage_path):
            # And create it if does not exists
            os.mkdir(getSysroot() + self.default_vstorage_path)

        # Create or Join MDS
        if self.need_mds():
            # Create MDS on SSD if it available
            mds_path = "/mnt/ssd1"
            if not os.path.exists(getSysroot() + mds_path):
                mds_path = "%s/%s-mds" % (self.default_vstorage_path, self.clustername)
                if not os.path.exists(getSysroot() + mds_path):
                    os.mkdir(getSysroot() + mds_path)
            mds_path += "/data"
            stdin=None
            parg = []
            initarg = []
            if self.need_mds_create():
                stdin=create_vstorage_input(self.clusterpw)
                parg = ["-P"]
                initarg = ["--init-cluster"]
            rc = execWithRedirectCl(self.vstorage_bin, self.common_opts + [ self.clustername,
                "make-mds" ] + initarg + [ "-a", "%s:%i" % (self.mds_ip, self.mds_port),
                "-r", mds_path ] + parg, root=getSysroot(), stdin=stdin,
                stdout = open(self.logfile, "a+"))
            if rc:
                log.error("MDS %s failed: %i" % ("create" if self.need_mds_create() else "join", rc))
                return False
            if self.need_mds_create():
                if not self.start_service("mdsd", [mds_path]):
                    return False
                # Create Container and VM dirs on clean cluster
                tmpdir = "/tmp/vstorage_m"
                os.mkdir(getSysroot() + tmpdir)
                execWithRedirectCl("/bin/mount", ["-t", "fuse.vstorage", "vstorage://%s" %
                    self.clustername , tmpdir], stdout = open(self.logfile, "a+"),
                    root=getSysroot())
                os.mkdir(getSysroot() + tmpdir + "/private")
                os.mkdir(getSysroot() + tmpdir + "/vmprivate")
                execWithRedirectCl("/bin/umount", [tmpdir], stdout = open(self.logfile, "a+"),
                    root=getSysroot())
                shutil.rmtree(getSysroot() + tmpdir)

        # Create CSes
        if self.cs_join:
            csnum = 0
            for part in self.vstorage_parts(storage):
                reg_timeout = self.reg_timeout + 1
                cspath = part.format.mountpoint + "/data"
                while (reg_timeout):
                    ssd_params = []
                    if len(ssd_caches) and self.cs_ssd:
                        ssh_cache_path = ssd_caches[0][0] + "/%s-cs%i-ssd" % (self.clustername, csnum)
                        if not os.path.exists(getSysroot() + ssh_cache_path):
                            os.mkdir(getSysroot() + ssh_cache_path)
                        ssd_params.append("--journal=%s" % ssh_cache_path)
                        ssd_params.append("--journal-size=%i" % ssd_caches[0][1])
                        ssd_params.extend(["-T", "ssd"])
                        ssd_caches.pop(0)
                    start_time = time.time()
                    rc = execWithRedirectCl(self.vstorage_bin, self.common_opts + [ self.clustername,
                        "make-cs", "-r", cspath, "--tier=%s" % part.format.get_tier() ] + ssd_params,
                        stdout = open(self.logfile, "a+"), root=getSysroot())
                    overal_time = time.time() - start_time
                    if rc:
                        log.error("CS join failed: %i, will try again", rc)
                        reg_timeout -= 1
                        if overal_time > 60:
                            log.warning("Time exceeded execution of command %s make-cs" % self.vstorage_bin)
                            continue
                        time.sleep(60 - overal_time)
                        continue
                    if not self.start_service("csd", [cspath]):
                        log.warning("Failed to start CSD service for %s mpoint" % cspath)
                    break
                if rc:
                    log.error("CS join failed: %i", rc)
                csnum += 1

        # Create Client
        if self.client:
            # Create client fstab and dir
            client_cache_param = []
            if len(ssd_caches) and self.client_ssd:
                client_cache_f = "%s/client_cache" % ssd_caches[0][0]
                if not os.path.exists(getSysroot() + client_cache_f):
                    open(getSysroot() + client_cache_f, 'w').close()
                client_cache_param.append("cache=%s" % client_cache_f)
                client_cache_param.append("cachesize=%i" % ssd_caches[0][1])
            self.client_dir = "%s/%s" % (self.default_vstorage_path, self.clustername)
            if not os.path.exists(getSysroot() + self.client_dir):
                os.mkdir(getSysroot() + self.client_dir)
            f = open("%s/etc/fstab" % getSysroot(), 'a+')
            client_cache_param.append("_netdev")
            f.write('\n' + "vstorage://%s %s fuse.vstorage %s 0 0" % (self.clustername,
                self.client_dir, ','.join(client_cache_param)) + '\n')
            f.close()
            os.symlink(self.client_dir + "/private", getSysroot() + "/vz/private")
            os.symlink(self.client_dir + "/vmprivate", getSysroot() + "/vz/vmprivate")

        # Check vstorage license
        if self.key:
            self.key = checkVstorageLicense(self.clustername, self.key)

        # Vstorage key
        if self.key:
            rc = execWithRedirectCl(self.vstorage_bin, self.common_opts + [ self.clustername,
                "load-license", "-p", self.key], stdout = open(self.logfile, "a+"),
                root=getSysroot())
            if rc:
                log.error("Failed to activate the %s license key: %s" % \
                     (self.clustername, self.key))
                self.key = ""

        # Enable High Availability
        if self.shaman and self.client:
            rc = execWithRedirectCl("/bin/mount", [self.client_dir], stdout = open(self.logfile, "a+"),
                root=getSysroot())
            if rc:
                log.error("Failed to mount %s Client to %s : %i" % (self.productname, self.client_dir, rc))
                return False
            rc = execWithRedirectCl(self.shaman_bin, [ "-c", self.clustername, "join", "-r" ],
                stdout = open(self.logfile, "a+"), root=getSysroot())
            if rc:
                log.error("Shaman join failed: %i...", rc)
                return False
            for role in ("VM:QEMU", "CT:VZ7"):
                rc = execWithRedirectCl(self.shaman_bin, [ "-c", self.clustername, "add-role", role ],
                    stdout = open(self.logfile, "a+"), root=getSysroot())
                if rc:
                    log.error("Failed to add shaman role %s: %i...", role, rc)
                    return False
            rc = execWithRedirectCl("/bin/umount", [self.client_dir], stdout = open(self.logfile, "a+"),
                root=getSysroot())
            if rc:
                log.error("Failed to umount %s Client to %s : %i" % (self.productname, self.client_dir, rc))

            self.configure_snmpd()

        # Stop
        self.stop_services()

        return True

class VzttUrlmap(commands.vztturlmap.VZ7_Vztturlmap):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.vztturlmap.VZ7_Vztturlmap.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.url_map_f = "/vz/template/conf/vztt/url.map"
        self.url_map_l = "/etc/vztt/url.map"

    def execute(self, storage, ksdata, instClass):
        if not self.urls:
            return True
        log.info("Changing %s..." % self.url_map_l)
        urlmapf = getSysroot() + self.url_map_f
        if not os.path.exists(urlmapf):
            return False
        f = open(urlmapf).read()
        for replace_strings_array in self.urls:
            if re.search(r'\\=', replace_strings_array):
                replace_strings = replace_strings_array.split('\=')
            else:
                replace_strings = replace_strings_array.split('=')
            if len(replace_strings) == 2:
                replace_what_re = '^' + re.sub( r'\$', '#?\$', replace_strings[0]) + '.*'
                fnew = ""
                line_not_found = True
                for url_map_line in f.splitlines():
                    url_map_line_unmodifyed = url_map_line
                    url_map_line = re.sub(replace_what_re, replace_strings[0] + "\t" +  replace_strings[1], url_map_line_unmodifyed)
                    if url_map_line_unmodifyed != url_map_line:
                        line_not_found = False
                    fnew += "%s\n" % url_map_line
                if line_not_found:
                    fnew += "%s\n" % (replace_strings[0] + "\t" +  replace_strings[1])
                f = fnew
        open(urlmapf, 'w').write(f)
        return True

class Cep(commands.cep.VZ7_CEP):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.cep.VZ7_CEP.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.guest_collector = "/usr/share/virtuozzo/cep-scripts/libvirt_guests.py"
        self.logfile = "/tmp/cep.log"
        self.service = "disp-helper"

    def trigger_disp_helper_service(self, mode):
        rc = execWithRedirectCl("/usr/bin/systemctl", [mode, self.service],
                stdout = open(self.logfile, "a+"),
                root=getSysroot())
        if rc:
            log.error("Failed to %s %s." % (mode, self.service))

    def execute(self, storage, ksdata, instClass):
        if not self.enabled_node:
            log.info("CEP is disabled.")
            self.trigger_disp_helper_service("disable")
            return

        log.info("Enabling CEP...")
        self.trigger_disp_helper_service("enable")

        if self.enabled:
            return

        log.info("Disabling CEP for guests...")
        try:
            os.chmod(getSysroot() + self.guest_collector, 0o644)
        except:
            log.error("Failed to disable %s guest collector." % self.guest_collector)

class Key(commands.key.RHEL5_Key):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.key.RHEL5_Key.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.logfile = "/tmp/license.log"
        self.vzlicload = "/usr/sbin/vzlicload"
        self.vzlicupdate = "/usr/sbin/vzlicupdate"
        self.vzlicview = "/usr/sbin/vzlicview"

    def checkLicense(self):
        if not self.key:
            return False

        if not os.access(getSysroot() + self.vzlicview, os.X_OK):
            log.error("An error occurred while installing the license! " \
                "%s is not found or not accessible" % self.vzlicview)
            return False

        buf = execWithCapture(self.vzlicview, ["-p", self.key ], root=getSysroot(),
            filter_stderr=True)
        log.info("Checking entered key %s" % self.key)
        if buf.lower().find("keyserver_host=") != -1:
            type_activation = True
            log.info("Entered key %s is an activation key." % self.key)
        elif buf.lower().find("status=\"valid\"") != -1:
            type_activation = False
            log.info("Entered key %s is a valid product key." % self.key)
        else:
            log.error("Entered key %s is not activation key or valid product key." % self.key)
            return False

        return True

    def processLicense(self):
        if not self.checkLicense():
            return

        if not os.access(getSysroot() +  self.vzlicload, os.X_OK):
            log.error("An error occurred while installing the license! " \
                "%s is not found or not accessible" % self.vzlicload)
            return

        log.info("Installing the new license...")
        if execWithRedirectCl(self.vzlicload, ["-p", self.key],
                 stdout = open(self.logfile, "a+"),
                 root=getSysroot()):
            log.error("An error occurred while installing the license. "
                                  "For details, see %s ." % os.path.basename(self.logfile))
            return

        log.info("The license has been installed successfully!")

    def execute(self, storage, ksdata, instClass):
        if instClass.is_free:
            return

        if self.key and not self.skip:
            self.processLicense()

class Prlnet(commands.prlnet.VZ7_Prlnet):
    def execute(self, storage, ksdata, instClass):
        if not self.used:
            return
        if not self.name:
            self.name = "Host-Only"
        log.info("Set IP addresses scope for network {self.name}, " \
            "start IP {self.ip_scope_start}, end IP {self.ip_scope_end} " \
            "ip/mask {self.ip} DHCP IP {self.dhcp_ip} " \
            "start IP6 {self.ip6_scope_start}, end IP6 {self.ip6_scope_end} " \
            "ip6/mask {self.ip6} DHCP IP6 {self.dhcp_ip6} " \
            "dhcp-server: {self.dhcp_server} dhcp6-server: {self.dhcp6_server}".format(**locals()))
        opts = ["net", "set", self.name]
        if self.ip_scope_start:
             opts.extend(["--ip-scope-start", self.ip_scope_start, \
                          "--ip-scope-end", self.ip_scope_end])
        if self.ip6_scope_start:
             opts.extend(["--ip6-scope-start", self.ip6_scope_start, \
                          "--ip6-scope-end", self.ip6_scope_end])
        if self.ip:
            opts.extend(["--ip", self.ip])
        if self.ip6:
            opts.extend(["--ip6", self.ip6])

        if self.dhcp_ip:
            opts.extend(["--dhcp-ip", self.dhcp_ip])
        if self.dhcp_ip6:
            opts.extend(["--dhcp-ip6", self.dhcp_ip6])

        if self.dhcp_server:
            opts.extend(["--dhcp-server", self.dhcp_server])
        if self.dhcp6_server:
            opts.extend(["--dhcp6-server", self.dhcp6_server])

        instClass.prl_disp_first_start(["/usr/bin/prlsrvctl " + " ".join(opts)])

class AutoPart(commands.autopart.RHEL7_AutoPart):
    def __init__(self, writePriority=100, *args, **kwargs):
        super(AutoPart, self).__init__(writePriority=writePriority, *args, **kwargs)

    def parse(self, args):
        retval = super(AutoPart, self).parse(args)

        ignored_keys = []
        if self.fstype:
            ignored_keys.append("--fstype")
        if self.encrypted:
            ignored_keys.append("--encrypted")
        if self.type:
            ignored_keys.append("--type")
        if not self.lvm:
            ignored_keys.append("--no-lvm")
        if ignored_keys:
            warnings.warn("The autopart kickstart key does not support: %s ." % \
                ", ".join(ignored_keys))

        return retval

    def execute(self, storage, ksdata, instClass):
        if not self.autopart:
            return

        DoVZAutopart(instClass, storage).execute()

class Readykernel(commands.readykernel.VZ7_Readykernel):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.readykernel.VZ7_Readykernel.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.command = "/usr/sbin/readykernel"
        self.autoupdate_cmd = ["autoupdate"]
        self.autoupdate_enable = self.autoupdate_cmd + ["enable"]
        self.autoupdate_disable = self.autoupdate_cmd + ["disable"]

    def execute(self, storage, ksdata, instClass):
        if instClass.is_free:
            return

        if not os.path.exists(getSysroot() + self.command):
            log.error("%s binary doesn't exist in sysroot" % self.command)
            return

        # Call utility
        rc = execInSysroot(self.command, self.autoupdate_enable if self.autoupdate else self.autoupdate_disable)
        if rc:
            log.error("Failed to execute %s command: %i", self.command, rc)
            return False
        return True

class Autoupdate(commands.autoupdate.VZ7_Autoupdate):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.autoupdate.VZ7_Autoupdate.__init__(self, writePriority=writePriority, *args, **kwargs)

    def execute(self, storage, ksdata, instClass):
        if instClass.is_free:
            return

        hour = int(random.random()*100) % 24
        minute = int(random.random()*100) % 60
        log.info("Trying to set autoupdate to %d %d..." % (hour, minute))
        if self.autoupdate:
            # Just in case, create folders if missing
            try:
                os.makedirs(getSysroot() + "/etc/systemd/system/vzautoupdate.timer.d")
            except:
                pass
            try:
                with open(getSysroot() + "/etc/systemd/system/vzautoupdate.timer.d/override.conf", "w") as f:
                   # Need to empty OnCalendar first to really override values from the unit
                   f.write("[Timer]\nOnCalendar=\nOnCalendar=*-*-* %d:%d:00" % (hour, minute))
            except Exception as e:
                log.error("Failed to edit autoupdate timer: %s" % str(e))
                return False
        else:
            rc = execInSysroot("/bin/systemctl", ["disable", "vzautoupdate.timer"])
            rc = execInSysroot("/bin/systemctl", ["disable", "vzautoupdate.service"])
        return True

class Container(commands.container.VZ7_Container):
    def place_precreated_cache(self, ksdata, instClass):
        if ksdata.nosfxtemplate.enabled or not instClass.is_vz:
            return True
        log.info("Trying to use precreated cache...")
        dcachedir = getSysroot() + instClass.dcachepath
        if not os.path.exists(dcachedir):
            os.makedirs(dcachedir)
        dcachepath = "%s/%s.%s" % (dcachedir, instClass.defostemplate,
            instClass.ostemplate_cache_suffix)

        return bring_file(instClass, "%s/%s.%s" % (instClass.isocachedir,
                instClass.defostemplate, instClass.ostemplate_cache_suffix), dcachepath) is not None

    def execute(self, storage, ksdata, instClass):
        # Bring the cache in any way
        self.place_precreated_cache(ksdata, instClass)
        for c in self.containerList:
            c.execute(storage, ksdata, instClass)

class ContainerData(commands.container.VZ7_ContainerData):
    def __init__(self, *args, **kwargs):
        commands.container.VZ7_ContainerData.__init__(self, *args, **kwargs)
        self.vzpkg = "/usr/sbin/vzpkg"
        self.vzctl = "/usr/sbin/vzctl"
        self.modules_list = "/etc/modules-load.d/vz.conf"
        self.to_bind_squashfs = ["/etc/vz", "/vz"]
        self.to_bind_normal = self.to_bind_squashfs + [self.vzctl, "/usr/sbin/vztactl", self.modules_list,
                        "/etc/vztt", self.vzpkg, "/usr/libexec/vztt_pfcache_xattr",
                        "/usr/libexec/vztt_checker_open.so", "/var/vz", "/usr/lib64/vztt",
                        "/usr/sbin/sgdisk", "/usr/bin/lz4", "/usr/libexec/libvzctl", "/usr/sbin/vzlist"]
        self.dep_binded = []
        self.to_unbind = []
        self.vz_script = "/usr/libexec/vz"
        self.modules_load = "/usr/lib/systemd/systemd-modules-load"
        # From vztt man page
        self.cache_exists_err = 29
        self.logfile = "/tmp/container.log"
        self.pre_scripts = []
        self.post_scripts = []
        self.container_id = str(uuid.uuid4())
        self.files_to_clean = [getSysroot() + '/etc/vz/prl_disp_disable_network_sync']
        self.additional_params = {}
        self.template_configs_path = "/vz/template/%s/config/os/default/"
        self.vz_conf = "/etc/vz/vz.conf"
        self.ksdata = None
        self.instClass = None
        self.vstorage_configure_backend_bin = "/usr/libexec/vstorage-ui-backend/bin"
        self.vstorage_ui_backend_bin = "/usr/libexec/vstorage-ui-backend/libexec"
        self.va_mn_ports = ['389', '4533', '4534', '4648', '4450', '4452', '80', '443', '1433', '22']
        self.firewall_cmd = "/usr/bin/firewall-offline-cmd"
        self.resolv_conf = "/etc/resolv.conf"
        self.private_path = "/vz/private/%s" % self.container_id
        self.root_path = "/vz/root/%s" % self.container_id
        self.container_scripts_path = "%s/scripts" % self.private_path
        self.config_path = "%s/ve.conf" % self.private_path
        self.ui_mount_script = {"file": "%s/premount" % self.container_scripts_path, "data" :
"#!/bin/bash\n" \
"\n" \
"FIREWALL_CMD=\"/usr/bin/firewall-cmd\"\n" \
"FIREWALL_OFFLINE_CMD=\"/usr/bin/firewall-offline-cmd\"\n" \
"AGENT_PORT=17514\n" \
"\n" \
". `dirname $0`/../ve.conf\n" \
"\n" \
"[ \"x$IP_ADDRESS\" = \"x\" ] && exit 0\n" \
"\n" \
"if [ \"x$1\" = \"x\" ]; then\n" \
"    ACTION=\"add\"\n" \
"elif [ \"$1\" = \"--remove\" ]; then\n" \
"    ACTION=\"remove\"\n" \
"else\n" \
"    exit 0\n" \
"fi\n" \
"\n" \
"for ctip in $IP_ADDRESS; do\n" \
"    [ \"${ctip#*:}\" = \"${ctip}\" ] && ip=\"ipv4\" || ip=\"ipv6\"\n" \
"    ARGS=\"--direct --$ACTION-rule $ip filter INPUT_direct 0 -i venet0 -s $ctip -m tcp -p tcp --dport $AGENT_PORT -j ACCEPT\"\n" \
"    $FIREWALL_CMD --permanent $ARGS && $FIREWALL_CMD --reload || $FIREWALL_OFFLINE_CMD $ARGS\n" \
"done\n" \
"\n" \
"exit 0\n"}
        self.ui_umount_script = {"file": "%s/umount" % self.container_scripts_path, "data" :
"#!/bin/bash\n" \
"\n" \
"`dirname $0`/premount --remove\n" \
"\n" \
"exit $?\n"}

    def get_missed_libs(self, path):
        if not os.path.isfile(path):
            return
        for x in [ x for x in execWithCapture("/usr/bin/ldd",
                [path]).split('\n') if x.endswith("not found")]:
            lib = "/usr/lib64/%s" % x.split()[0]
            self.dep_binded.append(lib)
            self.bind([lib])

    def bind_mount(self, what, to):
        rc = execWithRedirectCl("/bin/mount", [ '-o', 'bind', what, to ],
            stdout = open(self.logfile, 'a+'))
        if rc:
            log.error("Failed to bind %s" % to)
            return False
        return True

    def replace_file(self, what, to):
        if not os.path.exists(to):
            open(to, 'w').close()
        return self.bind_mount(what, to)

    def bind(self, to_bind):
        for path in to_bind:
            cpath = getSysroot() + path
            if not os.path.exists(path):
                if os.path.isdir(cpath):
                    os.mkdir(path)
                else:
                    open(path, 'w').close()

            if not self.bind_mount(cpath, path):
                return False

            # Get linked libraries
            self.get_missed_libs(path)

        return True

    def replace_tconfig(self, cname, data = []):
        tfile = tempfile.mktemp("_%s" % cname)
        if self.ostemplate.startswith("r-virtualization-7-x86_64"):
            destdir = self.template_configs_path % "r-virtualization/7/x86_64"
        else:
            destdir = self.template_configs_path % '/'.join(self.ostemplate.split('-'))
        if not os.path.exists(destdir):
            log.error("Template directory does not exist: %s" % destdir)
            return False
        destfile = destdir + cname
        with open(tfile, 'w') as f:
            for l in data:
                f.write(l + '\n')
        if not self.replace_file(tfile, destfile):
            return False
        self.to_unbind.append(destfile)
        self.files_to_clean.append(tfile)

        return True

    def setup_va_mn_vstorage_ui_plugin(self, addr):
        self.packages.append("va-cc-storage")
        return "sed -e \"s,<address>local</address>,<address>%s</address>,g\" -i /etc/opt/pva/cc/pvacc.conf" % addr

    def check_base_repos(self):
        # Should not happens, but
        if not self.instClass.base_repos and not self.instClass.base_mlists:
            log.error("Failed to get base repos and mirrorlists")
            return False

        return True

    def setup_params_va_mn(self):
        if self.special != "va-mn":
            return True

        if not self.check_base_repos():
            return False

        if not self.instClass.base_va_mn_repos and not self.instClass.base_va_mn_mlists:
            log.error("Failed to get MN repos and mirrorlists")
            return False

        if not self.replace_tconfig("mirrorlist", data = self.instClass.base_mlists + self.instClass.base_va_mn_mlists):
            return False

        if not self.replace_tconfig("repositories", data = self.instClass.base_repos + self.instClass.base_va_mn_repos):
            return False

        ct_post_scripts = []
        self.groups = [ "va-mn", "va-cc" ]
        self.packages = [ "va-mn-release" ]
        for c in [ c for c in self.ksdata.container.containerList if c.create and c.special == "vstorage-ui" and c.ip ]:
            ct_post_scripts.append(self.setup_va_mn_vstorage_ui_plugin(c.ip[0]))
            break
        if not ct_post_scripts and self.ksdata.registervstorageui.enabled and self.ksdata.registervstorageui.ui_addr:
            ct_post_scripts.append(self.setup_va_mn_vstorage_ui_plugin(self.ksdata.registervstorageui.ui_addr))

        self.additional_params["--ram"] = "1G"
        self.additional_params["--diskspace"] = "40G"

        for port in self.va_mn_ports:
            ct_post_scripts.append("%s --add-port %s/tcp" % (self.firewall_cmd, port))
        self.post_scripts = {"ct": ct_post_scripts, "host" : []}

        return True

    def setup_params_vstorage_ui(self):
        if self.special != "vstorage-ui":
            return True

        if not self.check_base_repos():
            return False

        if not self.replace_tconfig("mirrorlist", data = self.instClass.base_mlists):
            return False

        if not self.replace_tconfig("repositories", data = self.instClass.base_repos):
            return False

        self.packages_remove = [ "httpd" ]
        self.groups = [ "vstorage-ui-backend" ]

        self.additional_params["--ram"] = "2G"
        self.additional_params["--diskspace"] = "40G"
        self.post_scripts = {"ct": ["PGSETUP_INITDB_OPTIONS=\"-E UTF8\" postgresql-setup initdb",
            "echo \"%s\" | %s/configure-backend.sh -x venet0" % (self.rootpw, self.vstorage_configure_backend_bin),
            "sed -e \"s,^configure_firewall$,#configure_firewall,g\" -i %s/init-backend.sh" % self.vstorage_ui_backend_bin,
            "%s/init-backend.sh" % self.vstorage_ui_backend_bin,
            "sed -e \"s,^DefaultZone=.*,DefaultZone=trusted,g\" -i /etc/firewalld/firewalld.conf"],
            "host" : ["cp %s/etc/yum.repos.d/virtuozzo.repo %s/etc/yum.repos.d/" % (getSysroot(), self.root_path),]}

        return True

    def additional_bind(self):
        if not self.special:
            return True

        self.ostemplate = self.instClass.defostemplate

        return self.setup_params_va_mn() and self.setup_params_vstorage_ui()

    def unbind_p(self, path):
        rc = execWithRedirectCl("/bin/umount", [ path ],
            stdout = open(self.logfile, 'a+'))
        if rc:
            log.warning("Failed to unbind %s" % path)
            return False
        return True

    def unbind(self):
        for path in self.to_unbind:
            self.unbind_p(path)
        self.to_unbind = []
        for path in self.dep_binded + self.to_bind:
            if not self.unbind_p(path):
                continue
            try:
                if os.path.isfile(path):
                    os.unlink(path)
                else:
                    os.rmdir(path)
            except:
                log.warning("Failed to remove %s" % path)
        self.dep_binded = []

    def setup_vz(self, action):
        rc = execWithRedirectCl(getSysroot() + self.vz_script, [ action ],
            stdout = open(self.logfile, 'a+'))
        if rc:
            log.error("Failed to %s vz service" % action)
            return False
        # Disable vcmmd
        if action == "start":
            try:
                os.unlink("/usr/sbin/vcmmdctl")
            except:
                pass
        return True

    def load_modules(self):
        rc = execWithRedirectCl(self.modules_load, [ getSysroot() + self.modules_list ],
            stdout = open(self.logfile, 'a+'))
        if rc:
            log.error("Failed to load modules %s" % ' '.join(self.modules_list))
            return False
        return True

    def create_cache(self):
        # Check for cache exists
        if os.path.exists("%s/%s.%s" % (self.instClass.dcachepath,
                self.instClass.defostemplate, self.instClass.ostemplate_cache_suffix)):
            return True
        rc = execWithRedirectCl(self.vzpkg, [ "create", "cache", self.ostemplate ],
            stdout = open(self.logfile, 'a+'))
        if rc and rc != self.cache_exists_err:
            log.error("Failed to create cache %s" % self.ostemplate)
            return False
        return True

    def create_container(self):
        rc = execWithRedirectCl(self.vzctl, [ "create", self.container_id,
            "--ostemplate", self.ostemplate ], stdout = open(self.logfile, 'a+'))
        if rc:
            log.error("Failed to create container %s" % self.container_id)
            return False
        return True

    def do_container(self, what, args):
        rc = execWithRedirectCl(self.vzctl, [ what, self.container_id ] + args,
            stdout = open(self.logfile, 'a+'))
        if rc:
            if what == 'start' and rc == 32:
                log.warning("Container %s is already running, restarting!" % self.container_id)
                self.stop_container()
                return self.start_container()
            log.error("Failed to %s container %s" % (what, self.container_id))
            return False
        return True

    def start_container(self):
        return self.do_container("start", ["--wait"])

    def stop_container(self):
        # Case when no vzctl in chroot()
        if not os.path.exists(self.vzctl):
            return True
        self.do_container("stop", ["--fast"])
        # Do not fail here - processes will be killed
        return True

    def pkgsaction(self, action, arg, name, what):
        if not what:
            return True
        rc = execWithRedirectCl(self.vzpkg, [ action, self.container_id ] + arg + what,
            stdout = open(self.logfile, 'a+'))
        if rc:
            log.error("Failed to %s %ss %s to container %s" % (action, name, ' '.join(what),
                self.container_id))
            return False
        return True

    def groupinstall(self):
        return self.pkgsaction("install", ["-g"], "group", self.groups)

    def templateinstall(self):
        return self.pkgsaction("install", [], "template", self.templates)

    def packagesinstall(self):
        return self.pkgsaction("install", ["-p"], "package", self.packages)

    def packagesremove(self):
        return self.pkgsaction("remove", ["-p"], "package", self.packages_remove)

    def run_scripts(self, scripts):
        if not scripts:
            return
        for s in scripts["ct"]:
            rc = execWithRedirectCl(self.vzctl, ["exec", self.container_id, s], stdout = open(self.logfile, 'a+'))
            if rc:
                log.error("Failed to exec %s in container %s: %i" % (s, self.container_id, rc))
        for s in scripts["host"]:
            rc = execWithRedirectCl(s.split(' ')[0], s.split(' ')[1:], stdout = open(self.logfile, 'a+'))
            if rc:
                log.error("Failed to exec %s in HN context: %i" % (s, rc))

    def set_param(self, param, to):
        if not to:
            return True
        if type(to) == list:
            cmd = [ "set", self.container_id ]
            for p in to:
                cmd.append(param)
                cmd.append(p)
            cmd.append("--save")
        else:
            cmd =  [ "set", self.container_id, param, to, "--save" ]
        rc = execWithRedirectCl(self.vzctl, cmd, stdout = open(self.logfile, 'a+'))
        if rc:
            log.error("Failed to set container %s parameter %s to %s" % (self.container_id, param, to))
            return False
        return True

    def set_ip(self):
        for ip in self.ip:
            if not self.set_param("--ipadd", ip):
                return False
        return True

    def set_hostname(self):
        return self.set_param("--hostname", self.hostname)

    def set_rootpw(self):
        if not self.rootpw:
            return True
        return self.set_param("--userpasswd", "root:%s" % self.rootpw)

    def set_nameserver(self):
        if self.nameserver:
            if not self.set_param("--nameserver", self.nameserver):
                return False
        return True

    def set_name(self):
        if not self.special:
            return True
        if not self.set_param("--name", self.special):
            return False
        return True

    def set_params(self):
        log.info("Setting Container %s parameters." % self.container_id)
        if not self.set_ip() or not self.set_hostname() or not self.set_rootpw() or not self.set_nameserver() or not self.set_name():
            return False
        for p in self.additional_params:
            if not self.set_param(p, self.additional_params[p]):
                return False
        return True

    def set_ns_as_hn(self):
        if self.nameserver or not self.special:
            return False

        try:
            with open(self.resolv_conf, "r") as f:
                self.additional_params["--nameserver"] = []
                for l in [ l for l in f.readlines() if l.startswith("nameserver") ]:
                    self.additional_params["--nameserver"].append(re.sub("^nameserver *", "", l).rstrip('\n'))
        except:
            log.error("Failed to get nameservers from %s" % self.resolv_conf)
            return False
        return True

    def create_mount_scripts(self):
        if self.special != "vstorage-ui":
            return True

        try:
            for script in (self.ui_mount_script, self.ui_umount_script):
                with open(script["file"], "w") as f:
                    f.write(script["data"])
                    os.chmod(script["file"], 0o755)
        except:
            log.error("Failed to create mount script(s)")
            return False

        return True

    def set_special_type(self):
        if not self.special:
            return True

        try:
            with open(self.config_path, "a+") as f:
                if not f.read().endswith('\n'):
                    f.write('\n')
                f.write("TYPE=\"temporary\"\n")
                f.write("AUTOSTOP=\"stop\"\n")
        except Exception, e:
            log.error("Failed to set special type to Container %s: %s" % (self.container_id, e))
            return False

        return True

    def set_host_timezone(self):
        try:
            ct_localtime = self.root_path + "/etc/localtime"
            host_tz = "/usr/share/zoneinfo/Host"
            if os.path.exists(ct_localtime):
                os.unlink(ct_localtime)
            os.symlink(".." + host_tz, ct_localtime)
            shutil.copy(getSysroot() + "/etc/localtime", self.root_path + host_tz)
        except Exception, e:
            log.error("Failed to configure Container %s timezone: %s" % (self.container_id, e))
            return False
        return True

    def cleanup(self):
        self.stop_container()
        self.setup_vz("stop")
        self.unbind()
        for f in self.files_to_clean:
            try:
                os.unlink(f)
            except:
                pass
        self.files_to_clean = []

    def execute(self, storage, ksdata, instClass):
        if not self.create:
            return

        # For installation method detection
        self.ksdata = ksdata
        self.instClass = instClass

        if self.instClass.is_squashroot:
            self.to_bind = self.to_bind_squashfs
        else:
            self.to_bind = self.to_bind_normal

        if not self.ostemplate:
            with open(getSysroot() + self.vz_conf, "r") as f:
                for line in [ re.sub("\"", "-x86_64", re.sub("DEF_OSTEMPLATE=\"\.", "", l))
                        for l in f.readlines() if l.startswith("DEF_OSTEMPLATE") ]:
                    self.ostemplate = line
                    break

        # Set the same nameservers as for HN
        self.set_ns_as_hn()

        log.info("Container %s preparation started." % self.container_id)
        with open('/etc/vz/prl_disp_disable_network_sync', 'w') as f:
            pass
        try:
            os.makedirs(getSysroot() + '/etc/vz')
        except:
            pass
        with open(getSysroot() + '/etc/vz/prl_disp_disable_network_sync', 'w') as f:
            pass
        if self.load_modules() and \
                self.bind(self.to_bind) and \
                self.additional_bind() and \
                self.setup_vz("start") and \
                self.create_cache() and \
                self.create_container() and \
                self.set_params() and \
                self.start_container():
            self.run_scripts(self.pre_scripts)
            self.packagesremove()
            self.packagesinstall()
            self.templateinstall()
            self.groupinstall()
            self.run_scripts(self.post_scripts)
            self.create_mount_scripts()
            self.set_special_type()
            self.set_host_timezone()
            log.info("Container %s preparation successfully finished." % self.container_id)
        else:
            log.error("Failed to prepare Container %s ." % self.container_id)

        self.cleanup()

class RegisterVAagent(commands.register_vaagent.VZ7_RegisterVAagent):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.register_vaagent.VZ7_RegisterVAagent.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.logfile = "/tmp/register_vaagent.log"
        self.command = "/usr/sbin/vzagroup"
        self.setupagent = "/opt/pva/agent/bin/setupagent"
        self.agentconf = "/opt/pva/agent/bin/pva.conf"
        self.agentd = "/opt/pva/agent/bin/pvaagentd"
        self.prl_dispd = "/usr/sbin/prl_disp_service"
        self.prlsrvctl = "/usr/bin/prlsrvctl"
        self.containers = []
        self.timeout = 30
        self.prl_disp_timeout = self.timeout

    def cleanup(self):
        stop_containers(self.containers)

    def execute(self, storage, ksdata, instClass):
        if not self.enabled:
            return

        if not self.mn_addr:
            log.error("Failed to register va agent: no MN address given")
            return False
        if not self.rootpw:
            log.error("Failed to register va agent: no MN root password given")
            return False

        self.containers = start_containers(ksdata, "va-mn", self.mn_addr)

        rc = execWithRedirectCl(self.setupagent, [],
            stdout = open(self.logfile, 'a+'), root=getSysroot())
        if rc:
            log.error("Failed to setup va agent: %i" % rc)
            self.cleanup()
            return False

        # Fork here due to pvaagentd do not do it by yourself
        try:
            disp_pid = os.fork()
            if not disp_pid:
                os.chroot(getSysroot())
                os.chdir("/")
                os.execl(self.prl_dispd, self.prl_dispd)
            # Wait for prl-disp sockets
            while self.prl_disp_timeout > 0:
                rc = execWithRedirectCl(self.prlsrvctl, ["info"],
                    stdout = open(self.logfile, 'a+'), root=getSysroot())
                if rc == 0:
                    break
                time.sleep(1)
                self.prl_disp_timeout -= 1
                continue
            if rc:
                log.error("Failed to wait for prl_disp service during register va agent: %i" % rc)
                os.kill(disp_pid, signal.SIGKILL)
                os.waitpid(disp_pid, 0)
                self.cleanup()
                return False
            agent_pid = os.fork()
            if not agent_pid:
                os.chroot(getSysroot())
                os.chdir("/")
                os.execl(self.agentd, self.agentd, self.agentconf)
            addtogrouparg = "root:%s@%s" % (self.rootpw, self.mn_addr)
            # Wait for register
            while self.timeout > 0:
                rc = execWithRedirectCl(self.command, ["addToGroup", addtogrouparg ],
                    stdout = open(self.logfile, 'a+'), root=getSysroot())
                if rc == 0:
                    break
                time.sleep(1)
                self.timeout -= 1
                continue
            for pid in (agent_pid, disp_pid):
                os.kill(pid, signal.SIGKILL)
                os.waitpid(pid, 0)
            if rc:
                log.error("Failed to register va agent to %i: %s" % (rc, addtogrouparg))
                self.cleanup()
                return False
        except:
            log.error("Failed to start and register throught %s" % self.agentd)
        self.cleanup()
        return True

class RegisterVStorageUI(commands.register_vstorage_ui.VZ7_RegisterVStorageUI):
    def __init__(self, writePriority=100, *args, **kwargs):
        commands.register_vstorage_ui.VZ7_RegisterVStorageUI.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.ui_agent_conf = getSysroot() + "/etc/vstorage/vstorage-ui-agent.conf"
        self.host_id_conf = getSysroot() + "/etc/vstorage/host_id"
        self.logfile = "/tmp/register_vstorage_ui.log"
        self.command = "/usr/libexec/vstorage-ui-agent/bin/configure-agent.sh"
        self.agent_nginx_config = getSysroot() + "/etc/nginx/conf.d/vstorage-ui-agent.conf"
        self.is_virt = "true" if vt_supported() else "false"
        self.container_root_prefix = "/vz/root/%s"
        self.register_offline_cmd_f = "/tmp/register_offline_cmd_f"
        self.ui_agent_srv = "vstorage-ui-agent.service"
        self.init_agent_ct_srv = "init-agent-ct.service"
        self.init_agent_ct_config = getSysroot() + "/usr/libexec/vstorage-ui-agent/etc/ct_agent_cfg.init"
        self.register_offline_cmd_t = "#!/bin/bash\n" \
        "for i in `seq 1 10`; do systemctl is-active postgresql && break; sleep 60; done\n" \
        "psql -1c \"insert into nodes (id, host_id, host, cluster_id, is_virt) " \
        "values ('%s', '%s', '%s', null, '%s')\"\n" \
        "/usr/libexec/vstorage-ui-backend/libexec/gen-certificate.sh -a -i '%s'\n"
        self.correct_nginx_configs_f = "/tmp/correct_nginx_configs_f"
        self.correct_nginx_configs_t = "#!/bin/bash\n" \
        "/bin/sed -e \"s,^\([ \\t]*listen[ \\t]*\[*:*\]*:*80[; \\t]*\),#\\1,g\" -i /etc/nginx/nginx.conf\n" \
        "/bin/sed \"/^[ \\t]*listen[ \\t]*8888;.*\$/i\\\\\\tlisten 80;\" -i /etc/nginx/conf.d/vstorage-ui-backend.conf\n"
        # 5432
        self.ui_defport = 8888
        self.bound_ip = ""

    @property
    def node_id(self):
        node_id = ""
        try:
            with open(self.ui_agent_conf, "r") as f:
                for line in [ l for l in f.readlines() if re.match("NODE_ID( |\t)*=", l) ]:
                    node_id = re.sub("\'$", "", re.sub("^NODE_ID( |\t)*=( |\t)*\'", "", line)).rstrip()
                    break
        except:
            pass
        if not node_id:
            log.error("Failed to get node_id!")
        return node_id

    @property
    def host_id(self):
        host_id = ""
        try:
            with open(self.host_id_conf, "r") as f:
                host_id = f.read().rstrip()
        except:
            pass
        if not host_id:
            log.error("Failed to get host_id!")
        return host_id

    @property
    def bind_ip(self):
        if self.bound_ip:
            return self.bound_ip
        try:
            for ip in get_nonlocal_ip_list(with_prefix=True):
                if IPAddress(self.ui_ipv4) in IPNetwork(ip):
                    self.bound_ip = str(IPNetwork(ip).ip)
                    log.info("Host IP %s selected as agent source address for registration" % self.bound_ip)
                    return self.bound_ip
        except:
            log.error("Failed to assign %s MN IP for any HN interface" % self.ui_ipv4)
        log.warning("MN IP %s doesn't fit in any HN interface IP/Mask!" % self.ui_ipv4)
        ips = get_nonlocal_ip_list()
        if ips:
            self.bound_ip = ips[0]
            log.warning("IP %s selected as local agent IP" % self.bound_ip)
            return self.bound_ip
        log.error("No any public IP available!")
        return ""

    def create_script(self, ctid, name, data):
        try:
            fname = self.container_root_prefix % (ctid) + name
            with open(fname, "w") as f:
                f.write(data)
            os.chmod(fname, 0o755)
        except:
            log.error("Failed to create register script")

    def destroy_script(self, ctid, name):
        try:
            os.unlink(self.container_root_prefix % (ctid) + name)
        except:
            log.error("Failed to destroy register script")

    def enable_service(self, service):
        # Enable agent autostart
        rc = execWithRedirectCl("/usr/bin/systemctl", ["enable", service],
                stdout = open(self.logfile, 'a+'), root=getSysroot())
        if rc:
            log.error("Failed to enable %s : %i" % (service, rc))
            return False
        return True

    def enable_services(self):
        for srv in (self.ui_agent_srv, self.init_agent_ct_srv):
            if not self.enable_service(srv):
                return False
        return True

    def create_init_agent_ct_config(self):
        try:
            with open(self.init_agent_ct_config, "w") as f:
                f.write("MN_IP_ADDR=\"%s\"\n" % self.ui_ipv4)
                f.write("MN_PORT=\"%i\"\n" % self.ui_defport)
        except:
            log.error("Failed to create config %s" % self.init_agent_ct_confi)
            return False
        return True

    def copy_ca_certs(self, veroot):
        nginx_ssl_dir = getSysroot() + "/etc/nginx/ssl/"
        if not os.path.exists(nginx_ssl_dir):
            try:
                os.mkdir(nginx_ssl_dir)
            except:
                log.error("Failed to create %s" % nginx_ssl_dir)
                pass
        for cert in ("crt", "key"):
            for cname, ctarg in (("ca", "ca"), ("agent-%s" % self.bind_ip, "server")):
                try:
                    shutil.copy(veroot + "/usr/libexec/vstorage-ui-backend/ca/%s.%s" % (cname, cert), nginx_ssl_dir + "/%s.%s" % (ctarg, cert))
                except:
                    log.error("Failed to copy %s" % cert)
                    return False
        return True

    def register_in_container(self, ksdata):
        containers_to_start = start_containers(ksdata, "vstorage-ui", self.ui_addr)

        self.register_offline_cmd = self.register_offline_cmd_t % (self.node_id, self.host_id, self.bind_ip, self.is_virt, self.bind_ip)

        if not containers_to_start:
            log.error("No VStorage UI Container available")
            return False

        if not self.node_id or not self.host_id or not self.bind_ip:
            log.error("Not enough parameters to register UI agent!")
            return False

        # Set ssl on by default - new installation
        try:
            with open(self.agent_nginx_config, "r+") as f:
                data = ""
                for l in f.readlines():
                    data += re.sub("#ssl off;", "#ssl on;", l)
                f.seek(0)
                f.truncate()
                f.write(data)
        except Exception, e:
            log.error("Failed to enable ssl for %s: %s" % (self.agent_nginx_config, e))
            return False

        commands_container = ["systemctl start postgresql", "su - vstoradmin -c \"%s\"" % self.register_offline_cmd_f, self.correct_nginx_configs_f]
        scripts_container = ((self.register_offline_cmd_f, self.register_offline_cmd), (self.correct_nginx_configs_f, self.correct_nginx_configs_t))
        for c in containers_to_start:
            for p in scripts_container:
                self.create_script(c.container_id, p[0], p[1])
            for cmd in commands_container:
                rc = execWithRedirectCl(c.vzctl, ["exec", c.container_id, cmd], stdout = open(self.logfile, 'a+'))
                if rc != 0:
                    log.error("Failed to exec %s in container %s: %i" % (cmd, c.container_id, rc))
                    break
            for p in scripts_container:
                self.destroy_script(c.container_id, p[0])
            if rc == 0 and not self.copy_ca_certs(c.root_path):
                rc = 1
                log.error("Failed to copy CA certificates from container %s" % c.container_id)

        stop_containers(containers_to_start)
        return rc == 0

    def execute(self, storage, ksdata, instClass):
        if not self.enabled:
            return

        # Get IPs here, not available at init stage
        # storage register script accepts ipv4 only :(
        try:
            self.ui_ipv4 = socket.getaddrinfo(self.ui_addr, 80, socket.AF_INET)[0][4][0]
        except Exception, e:
            log.error("Failed to get ipv4 for %s: %s" % (self.ui_addr, e))
            return False

        if not self.token:
            return self.register_in_container(ksdata) and self.create_init_agent_ct_config() and self.enable_services()

        input_data = create_vstorage_input(self.token)
        rc = execWithRedirect(self.command, ["-x", "%s:%i" % (self.ui_ipv4, self.ui_defport) ],
            stdout = open(self.logfile, 'a+'), root=getSysroot(), stdin=input_data)
        if rc:
            log.error("Failed to register vstorage-ui agent to %s with token %s: %i" % (self.ui_addr, self.token, rc))
            return False

        return True
